# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import copy, service
import hub, events, requests
import network

class player(service.service):
    def __init__(self, id, playlistid, networklocation):
        service.service.__init__(self, "remote player")
        self.id = id
        self.playlistid = playlistid
        self.networklocation = networklocation
        self.remoteplayerid = "main"

        self.networkchannel = network.clientchannel(self.networklocation)
        #self.networkchannel.transmit(events.updatesong)
        #self.networkchannel.transmit(events.updatealbum)
        #self.networkchannel.transmit(events.updateartist)
        #self.networkchannel.transmit(events.playlistaddsong)
        self.networkchannel.subscribe(events.playbackinfochanged, self.playbackinfochanged)

        # for playlists
        self.networkchannel.subscribe(events.playlistchanged, self.playlistchanged)
        self.networkchannel.start()

        # provide player service
        self.channel.subscribe(events.playerevent, self.playerevent)
        
        # we also provide a playlist service
        self.channel.subscribe(events.playlistevent, self.playlistevent)
        self.channel.supply(requests.playlistgetcontents, self.playlistgetcontents)
        
    # event handler

    def playbackinfochanged(self, event):
        # this is called by the networkchannel, so it should be (and is) thread safe
        event.playerid = self.id
        hub.notify(event)

    def playerevent(self, event):
        # we have to copy the event, because another thread may also access it
        event = copy.copy(event)
        event.playerid = self.remoteplayerid
        self.networkchannel.notify(event)

    def playlistevent(self, event):
        self.networkchannel.notify(event)

    def playlistchanged(self, event):
        hub.notify(event)

    # request handler

    def playlistgetcontents(self, request):
        return self.networkchannel.request(request)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path
import random
import time
import pickle

import config
import events, hub, requests
import item
import log
import service
import encoding

_counter = 0

class playlistitem:
    """ wrapped song with the two additional attributes played and id

    - id:     unique id for playlist item (note that the same song
              can be present more than once in the playlist)
    - played: has playlist item already been played
    - playstarttime: time at which song has been played or None if played is False
    """

    def __init__(self, song, played=False, playstarttime=None):
        global _counter
        self.song = song
        self.played = played
        self.playstarttime = playstarttime
        self.id = _counter
        _counter += 1

    def __repr__(self):
        return "playlistitem: id=%s" % `self.id`

    def getid(self):
        return self.id

    def getinfo(self):
        return self.song.getinfo()

    def getinfolong(self):
        return self.song.getinfolong()

    def markplayed(self):
        self.played = True
        self.playstarttime = time.time()

    def markunplayed(self):
        self.played = False

    def hasbeenplayed(self):
        return self.played


def initplaylist(id, playerid, songdbid):
    """initialize playlist service corresponding to player with playerid
    """
    playlist(id, playerid, songdbid).start()


class playlist(service.service):
    """manage playlist for a single player, which can be accessed 
    by multipled users"""

    def __init__(self, id, playerid, songdbid):
        service.service.__init__(self, "playlist")
        self.id = id
        # each playlist service is identified by the corresponding player
        self.playerid = playerid
        self.songdbid = songdbid
        self.items = []
        self.ttime = 0
        self.ptime = 0
        self.playingitem = None
        self.logfilename = config.general.logfile
        self.autoplaymode = config.general.autoplaymode

        self.channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
        self.channel.subscribe(events.playerstop, self.playerstop)
        self.channel.subscribe(events.playlistaddsongs, self.playlistaddsongs)
        self.channel.subscribe(events.playlistaddsongtop, self.playlistaddsongtop)
        self.channel.subscribe(events.playlistdeletesong, self.playlistdeletesong)
        self.channel.subscribe(events.playlistmovesongup, self.playlistmovesongup)
        self.channel.subscribe(events.playlistmovesongdown, self.playlistmovesongdown)
        self.channel.subscribe(events.playlistclear, self.playlistclear)
        self.channel.subscribe(events.playlistdeleteplayedsongs,
                               self.playlistdeleteplayedsongs)
        self.channel.subscribe(events.playlistreplay, self.playlistreplay)
        self.channel.subscribe(events.playlistsave, self.playlistsave)
        self.channel.subscribe(events.playlistshuffle, self.playlistshuffle)
        self.channel.subscribe(events.playlisttoggleautoplaymode, self.playlisttoggleautoplaymode)
        self.channel.subscribe(events.playlistplaysong, self.playlistplaysong)
        self.channel.subscribe(events.songchanged, self.songchanged)

        self.channel.supply(requests.playlist_requestnextsong, self.playlist_requestnextsong)
        self.channel.supply(requests.playlistgetcontents, self.playlistgetcontents)

        # try to load dump from prior crash, if existent
        if config.general.dumpfile:
            try:
                if os.path.isfile(config.general.dumpfile):
                    self.load()
                    os.unlink(config.general.dumpfile)
            except:
                pass

    def append(self, item):
        self.ttime += item.song.length
        if item.hasbeenplayed():
            self.ptime += item.song.length
        self.items.append(item)

    def insert(self, index, item):
        self.ttime += item.song.length
        if item.hasbeenplayed():
            self.ptime += item.song.length
        self.items.insert(index, item)

    def __delitem__(self, index):
        item = self.items[index]
        self.ttime -= item.song.length
        if item.hasbeenplayed():
            self.ptime -= item.song.length
        self.items.__delitem__(index)
        self._updateplaystarttimes()

    # all methods starting with an underscore may modify the playlist but leave
    # it up to the caller to announce this change via an playlistchanged event

    def _searchnextitem(self):
        """return playlistitem which has to be played next or None"""
        for i in range(len(self.items)):
            if not self.items[i].hasbeenplayed():
                return self.items[i]
        return None

    def _logplay(self, item):
        if self.logfilename:
            logfile = open(self.logfilename, "a")
            logfile.write("%s: %s\n" % (time.asctime(), encoding.encode_path(item.song.url)))
            logfile.close()

    def _updateplaystarttimes(self):
        # TODO: take crossfading time into account
        if self.playingitem:
            playstarttime = self.playingitem.playstarttime + self.playingitem.song.length
        else:
            playstarttime = time.time()
        for item in self.items:
            if not item.hasbeenplayed():
                item.playstarttime = playstarttime
                playstarttime += item.song.length

    def _playitem(self, item):
        """ check for a song abortion, register song as being played
        and update playlist information accordingly"""

        if not item.hasbeenplayed():
            self.ptime += item.song.length
        self.playingitem = item
        item.markplayed()
        self._updateplaystarttimes()
        self._logplay(item)

    def _playnext(self):
        """ mark next item from playlist as played and as currently playing and return
        corresponding song"""
        nextitem = self._searchnextitem()
        if nextitem:
            self._playitem(nextitem)
            return nextitem
        else:
            return None

    def _playprevious(self):
        """ mark next item from playlist as played & currently playing and return
        corresponding song"""

        # start either from the currently playing song, or if no song
        # is currently played, the first unplayed song in the playlist...
        #
        if self.playingitem:
            currentitem = self.playingitem
        else:
            currentitem = self._searchnextitem()

        if currentitem:
            # ... and go back one song
            i = self.items.index(currentitem)
            if i == 0:
                return
            self._markunplayed(currentitem)
            item = self.items[i-1]
            self._playitem(item)
            return item

    def _clear(self):
        self.items = []
        self.ptime = 0
        self.ttime = 0
        self.playingitem = None

    def _deleteplayedsongs(self):
        for i in range(len(self.items)-1,-1,-1):
            if self.items[i].hasbeenplayed() and self.items[i] != self.playingitem:
                del self[i]

    def _checksong(self, song):
        # it is ok if the song is contained in a local song database, so we first
        # check whether this is the case.
        # XXX make this behaviour configurable?
	stats = hub.request(requests.getdatabasestats(song.songdbid))
        if isinstance(song, item.song):
            if stats.type == "local":
                return song

	return song

        # XXX do we really need this
        # currently it does not work anymore
        if os.path.isfile(song.path):
            # first we try to access the song via its filesystem path
            return hub.request(requests.queryregistersong(self.songdbid, song.path))

        if song.artist != dbitem.UNKNOWN and song.album != dbitem.UNKNOWN:
            # otherwise we use the artist and album tags and try to obtain the song via
            # the database
            songs = hub.request(requests.getsongs(self.songdbid,
                                                  artist=song.artist, album=song.album))
            for asong in songs:
                if asong.title == song.title:
                    return asong

        # song not found
        # XXX start transmitting song
        return

    def _addsongs(self, songs):
        """add songs to end of playlist"""
        for song in songs:
            if song:
                song = self._checksong(song)
                if song:
                    self.append(playlistitem(song))
        self._updateplaystarttimes()

    def _markunplayed(self, item):
        """ mark song unplayed and adjust playlist information accordingly """
        if item.hasbeenplayed():
            self.ptime -= item.song.length
            item.markunplayed()
            self._updateplaystarttimes()

    def _markallunplayed(self):
        """ mark all songs in playlist as not having been played """
        for item in self.items:
            self._markunplayed(item)

    # convenience method for issuing a playlistchanged event

    def notifyplaylistchanged(self):
        hub.notify(events.playlistchanged(self.items, self.ptime, self.ttime, self.autoplaymode, self.playingitem))

    # statusbar input handler

    def saveplaylisthandler(self, name, key):
        name = name.strip()
        if key == ord("\n") and name != "" and self.items:
            songs = [item.song for item in self.items if item.song.songdbid == self.songdbid ]
            hub.notify(events.add_playlist(self.songdbid, name, songs))

    def _locatesong(self, id):
        """ locate position of item in playlist by id """
        for item, i in zip(self.items, range(len(self.items))):
            if item.id == id:
                return i
        else:
            return None

    def dump(self):
        """ write playlist to dump file """
        if self.playingitem:
            self.playingitem.markunplayed()
        # self._deleteplayedsongs()
        self.notifyplaylistchanged()
        dumpfile = open(config.general.dumpfile, "w")
        pickle.dump(self.items, dumpfile)

    def load(self):
        """ load playlist from file """
        dumpfile = open(config.general.dumpfile, "r")
        self._clear()
        for item in pickle.load(dumpfile):
            # We have to be careful here and not use the playlist item
            # stored in the dump file directly, since its id and the
            # global _counter variable are not in accordance. Besides that
            # the playstarttime information stored in the pickle is incorrect.
            # We thus have to create a new playlistitem.
            newplaylistitem = playlistitem(item.song, item.played, item.playstarttime)
            self.append(newplaylistitem)
        self._updateplaystarttimes()

    # event handlers

    def playbackinfochanged(self, event):
       # We are only interested in the case of the player having been stopped due to
       # no more song being left in the playlist.
       if event.playbackinfo.isstopped():
           self.playingitem = None
           self.notifyplaylistchanged()

    def playerstop(self, event):
        # Mark the currently playing song as unplayed again when the
        # player has been stopped manually. Note that the handling of
        # this event is potentially racy with the playbackinfochanged
        # event, but the ordering of the events in our event channel
        # should prevent any problems.
        if event.playerid == self.playerid:
            if self.playingitem:
                self._markunplayed(self.playingitem)
                self.playingitem = None
                self.notifyplaylistchanged()

    def playlistaddsongs(self, event):
        self._addsongs(event.songs)
        self.notifyplaylistchanged()

    def playlistaddsongtop(self, event):
        if event.song:
            song = self._checksong(event.song)
            if song:
                newitem = playlistitem(song)
                for i in range(len(self.items)):
                    if not self.items[i].hasbeenplayed():
                        self.insert(i, newitem)
                        break
                else:
                    self.append(newitem)
                self._playitem(newitem)
                self._updateplaystarttimes()
                hub.notify(events.playerplaysong(self.playerid, newitem))
                self.notifyplaylistchanged()

    def playlistdeletesong(self, event):
        i = self._locatesong(event.id)
        if i is not None:
            del self[i]
            self.notifyplaylistchanged()

    def playlistmovesongup(self, event):
        i = self._locatesong(event.id)
        if i is not None and i > 0:
            self.items[i-1], self.items[i] = self.items[i], self.items[i-1]
            self._updateplaystarttimes()
            self.notifyplaylistchanged()

    def playlistmovesongdown(self, event):
        i = self._locatesong(event.id)
        if i is not None and i<len(self.items)-1:
            self.items[i], self.items[i+1] = self.items[i+1], self.items[i]
            self._updateplaystarttimes()
            self.notifyplaylistchanged()

    def playlistclear(self, event):
        self._clear()
        self.notifyplaylistchanged()

    def playlistdeleteplayedsongs(self, event):
        self._deleteplayedsongs()
        self.notifyplaylistchanged()

    def playlistreplay(self, event):
        self._markallunplayed()
        self.notifyplaylistchanged()

    def playlistsave(self, event):
        hub.notify(events.requestinput(_("Save playlist"),
                                       _("Name:"),
                                       self.saveplaylisthandler))

    def playlistshuffle(self, event):
        random.shuffle(self.items)
        self._updateplaystarttimes()
        self.notifyplaylistchanged()

    def playlisttoggleautoplaymode(self, event):
        if self.autoplaymode == "off":
            self.autoplaymode = "repeat"
        elif self.autoplaymode == "repeat":
            self.autoplaymode = "random"
        else:
            self.autoplaymode = "off"
        self.notifyplaylistchanged()

    def playlistplaysong(self, event):
        i = self._locatesong(event.id)
        self._playitem(self.items[i])
        self.notifyplaylistchanged()
        hub.notify(events.playerplaysong(self.playerid, self.items[i]))

    def songchanged(self, event):
        # check whether one of our playlist items is affected by the change
        # in the songdb
        for item in self.items:
            if item.song == event.song and item.song.id==event.songdbid:
                item.song = event.song
                self._updateplaystarttimes()
                self.notifyplaylistchanged()

    def quit(self, event):
        service.service.quit(self, event)
        self.dump()

    #
    # request handler
    #

    def playlist_requestnextsong(self, request):
        if request.playlistid != self.id:
            raise hub.DenyRequest
        if not request.previous:
            nextitem = self._playnext()
            if not nextitem:
                if self.autoplaymode == "random":
                    # add some randomly selected song to the end of the playlist
                    randomsongs = hub.request(requests.getsongs("main", random=True))
                    if randomsongs:
                        self._addsongs(randomsongs[0:1])
                        nextitem = self._playnext()
                elif self.autoplaymode == "repeat":
                    self._markallunplayed()
                    nextitem = self._playnext()
        else:
            nextitem = self._playprevious()
        self.notifyplaylistchanged()
        return nextitem

    def playlistgetcontents(self, request):
        return self.items, self.ptime, self.ttime, self.autoplaymode, self.playingitem



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2005, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses, os, os.path, signal, imp, sys

##############################################################################
# gettext initialization.
##############################################################################

# We have to initialize gettext very early, before importing our
# modules. Assume that the locales lie in same dir as this
# module. This may not be the case, if the .mo files are installed at
# their proper location.

try:
    import gettext
    locallocaledir = os.path.join(os.path.dirname(sys.argv[0]), "../locale")
    gettext.install("PyTone", locallocaledir, unicode=True)
except:
    # Disable localization if there is any problem with the above.
    # This works around a problem with Python 2.1
    import __builtin__
    __builtin__.__dict__['_'] = lambda s: s

##############################################################################
# locale initialization
##############################################################################

import locale
locale.setlocale(locale.LC_ALL, '')

##############################################################################
# create .pytone dir in user home
##############################################################################
try:
    os.mkdir(os.path.expanduser("~/.pytone"))
    log.info(_("created PyTone directory ~/.pytone"))
except OSError, e:
    if e.errno!=17:
        raise

##############################################################################
# process commandline options and read config file
##############################################################################

import config
# process the command line first, because a different location for the
# config file may be given there
config.processcommandline()
config.processconfig()

# now that the configuration has been read, we can imort the log
# module and initialize the debug file if necessary
import log
log.initdebugfile(config.general.debugfile)

# log version (but do this after the debug file has been initialized,
# such that the version gets included there)
import version
log.info(_("PyTone %s startup") % version.version)

import errors
import mainscreen
import helper
import hub, events
import services.songdb
import services.player
import services.timer

# Uncomment the next line, if you want to experiment a little bit with
# the number of bytecode instructions after which a context switch
# occurs.

# sys.setcheckinterval(250)


##############################################################################
# start various services
##############################################################################

# catch any exceptions during service startup to be able to shut down
# all already running services when something goes wrong
try:
    # timer service. We start this first so that other services can register
    # there periodic events with this service
    services.timer.timer().start()

    # Determine plugins specified in the config file and read their config.
    # The result goees into a list of tuples (pluginmodule, pluginconfig).
    plugins = []

    userpluginpath = os.path.expanduser("~/.pytone/plugins/")
    cwd = os.path.abspath(os.path.dirname(sys.argv[0]))
    globalpluginpath = os.path.join(cwd, "plugins")
    pluginpath = [userpluginpath, globalpluginpath]

    for name in config.general.plugins:
        try:
            # We use imp.find_module to narrow down the plugin search path
            # to the two possible locations. Setting sys.path correspondingly
            # would not work, however, since then the plugin could not
            # import its needed modules. 
            fp, pathname, description = imp.find_module(name, pluginpath)
            pluginmodule = imp.load_module(name, fp, pathname, description)
            # 
            # process configuration of plugin
            pluginconfig = pluginmodule.config
            if pluginconfig is not None:
                config.readconfigsection("plugin.%s" % name, pluginconfig)
                config.finishconfigsection(pluginconfig)
                pluginconfig = pluginconfig()
            plugins.append((pluginmodule, pluginconfig))
        except Exception, e:
             log.error(_("Cannot load plugin '%s': %s") % (name, e))
             log.debug_traceback()

    # initialize song database manager and start it immediately so
    # that it can propagate quit events in case something goes wrong
    # when setting up the databases
    songdbmanager = services.songdb.songdbmanager()
    songdbmanager.start()

    # song databases
    songdbids = []
    for songdbname in config.database.getsubsections():
        try:
            songdbid = songdbmanager.addsongdb(songdbname, config.database[songdbname])
            if songdbid:
                songdbids.append(songdbid)
        except Exception, e:
            log.error("cannot initialize db %s: %s" % (id, e))

    if not songdbids:
        # raise last configuration error
        raise 

    # network service
    if config.network.enableserver:
        import network
        network.tcpserver(config.network.bind, config.network.port).start()
    if config.network.socketfile:
        import network
        network.unixserver(os.path.expanduser(config.network.socketfile)).start()

    # Now that the basic services have been started, we can initialize
    # the players. This has to be done last because the players
    # immediately start requesting a new song
    playerids = [services.player.initplayer("main", config.player.main),
                 services.player.initplayer("secondary", config.player.secondary)]

except:
    # if something goes wrong, shutdown all already running services
    hub.notify(events.quit(), 100)
    raise

##############################################################################
# basic curses library setup...
##############################################################################

def cursessetup():
    # Initialize curses library
    stdscr = curses.initscr()

    # Turn off echoing of keys
    curses.noecho()

    # In keypad mode, escape sequences for special keys
    # (like the cursor keys) will be interpreted and
    # a special value like curses.KEY_LEFT will be returned
    stdscr.keypad(1)

    # allow 8-bit characters to be input
    curses.meta(1)

    # enter raw mode, thus disabling interrupt, quit, suspend and flow-control keys
    curses.raw()

    # wait at maximum for 1/10th of seconds for keys pressed by user
    curses.halfdelay(1)

    if config.general.colorsupport == "auto":
        # Try to enable color support
        try:
            curses.start_color()
        except:
            log.warning("terminal does not support colors: disabling color support")

        # now check whether color support really has been enabled
        if curses.has_colors():
            config.configcolor._colorenabled = 1
    elif config.general.colorsupport == "on":
        curses.start_color()
        config.configcolor._colorenabled = 1

    # Check for transparency support of terminal
    # use_default_colors(), which will be integrated in python 2.4.
    # Before that happens we try to use our own cursext c-extension
    try:
        curses.use_default_colors()
        config.configcolor._colors["default"] = -1
    except:
        try:
            import cursext
            if cursext.useDefaultColors():
                config.configcolor._colors["default"] = -1
            else:
                log.warning("terminal does not support transparency")
        except:
            log.warning("transparency support disabled because cursext module is not present")

    # try disabling cursor
    try:
        curses.curs_set(0)
    except:
        log.warning("terminal does not support disabling of cursor")

    if config.general.mousesupport:
        # enable all mouse events
        curses.mousemask(curses.ALL_MOUSE_EVENTS)

    # redirect stderr to /dev/null (to prevent spoiling the screen
    # with libalsa messages). This is not really nice but at the moment there
    # is no other way to get rid of this nuisance.
    dev_null = file("/dev/null", 'w')
    os.dup2(dev_null.fileno(), sys.stderr.fileno())

    return stdscr

##############################################################################
# ... and cleanup
##############################################################################

def cursescleanup():
    # restore terminal settings
    try:
        stdscr.keypad(0)
        curses.echo()
        curses.nocbreak()
        curses.endwin()
    except:
        pass


##############################################################################
# signal handler
##############################################################################

def sigtermhandler(signum, frame):
    # shutdown all running threads
    hub.notify(events.quit(), 100)

signal.signal(signal.SIGTERM, sigtermhandler)

##############################################################################
# setup main screen (safety wrapped)
##############################################################################

try:
    stdscr = cursessetup()
    # set m to None as marker in case that something goes wrong in the
    # mainscreen.mainscreen constructor
    m = None
    m = mainscreen.mainscreen(stdscr, songdbids, playerids, plugins)
    m.run()
except:
    cursescleanup()

    # shutdown all other threads
    hub.notify(events.quit(), 100)

    helper.print_exc_plus()
    raise
else:
    cursescleanup()

    # shutdown all other threads
    hub.notify(events.quit(), 100)



### Terminal title plugin for pytone
### """"""""""""""""""""""""""""""""
### This plugin updates terminal and screen titles whenever pytone
### changes to a new song. You can install this plugin by simply
### copying it to your pytone system-wide plugin directory 
### (eg. /usr/share/pytone/plugins or ~/pytone/plugins/). And
### enable it by adding it to the plugins list in the general
### section, like:
###
###     [general]
###     plugins = termtitle
###
### You can customise the title by adding the following to
### /etc/pytonerc or ~/.pytone/pytonerc
###
###     [plugin.termtitle]
###     songformat = pytone: %(title)s - %(artist)s - %(length)s
###
### You can find the latest release at:
###
###     http://dag.wieers.com/home-made/pytone/
###
### If you have improvements or changes to this plugin, please
### send them to the pytone mailinglist and include me as well:
###
###     pytone-users@luga.de, Dag Wieers <dag@wieers.com>

import events, encoding, plugin, config, sys, os, re

class config(config.configsection):
    songformat = config.configstring('%(artist)s - %(title)s (%(length)s)')

class plugin(plugin.plugin):
    def init(self):
        if re.compile('(screen|xterm*)').match(os.getenv('TERM')):
            self.channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
            self.previoussong = None
            self.previouscross = False
        else:
            raise RuntimeError("Terminal type not supported")

    def playbackinfochanged(self, event):
        if event.playbackinfo.song and event.playbackinfo.song != self.previoussong or event.playbackinfo.iscrossfading() != self.previouscross:
            self.changetermtitle(event)
            self.previoussong = event.playbackinfo.song
            self.previouscross = event.playbackinfo.iscrossfading()

    def changetermtitle(self, event):
        prefix = (event.playbackinfo.iscrossfading() and '-> ' or '')
        song = encoding.encode(event.playbackinfo.song.format(self.config.songformat, safe=True))
        sys.stdout.write('\033]0;' + prefix + song + '\007')

# vim:ts=4:sw=4



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import copy, gc, math, random, service, time
import config
import events, hub, requests
import metadata
import item
import log

# helper function for the random selection of songs


#
# a collection of statistical information
#

class songdbmanagerstats:
    def __init__(self, songdbsstats, requestcachesize, requestcachemaxsize,
                 requestcacherequests, requestcachehits, requestcachemisses):
        self.songdbsstats = songdbsstats
        self.requestcachesize = requestcachesize
        self.requestcachemaxsize = requestcachemaxsize
        self.requestcacherequests = requestcacherequests
        self.requestcachehits = requestcachehits
        self.requestcachemisses = requestcachemisses

#
# the song database manager class
#

class songdbmanager(service.service):
    """ song database manager

    The song database manager receives database events and requests, passes them on to
    the various databases, collects the results and delivers them back to the caller.
    - Results of database requests are cached.
    - dbitem.song instances are wrapped in item.song instances which also contain the
      id of the database where the song is stored.
    - Random song selections are handled.

    """

    def __init__(self):
        service.service.__init__(self, "songdb manager")

        # hub for the various song databases
        self.songdbhub = hub.hub()

        # list of registered songdbs
        self.songdbids = []

        # result cache containing a mapping hash(request) -> (request, result, lastaccess)
        self.requestcache = {}
        # maximal number of objects referred by request cache
        self.requestcachemaxsize = config.database.requestcachesize
        # cache use statististics
        self.requestcachehits = 0
        self.requestcachemisses = 0
        # current number of objects referred to by items in result cache
        self.requestcachesize = 0 

        # we are a database service provider...
        self.channel.supply(requests.dbrequestsingle, self.dbrequestsingle)
        self.channel.supply(requests.dbrequestsongs, self.dbrequestsongs)
        self.channel.supply(requests.dbrequestlist, self.dbrequestlist)
        self.channel.supply(requests.getdatabasestats, self.getdatabasestats)
        self.channel.supply(requests.getnumberofsongs, self.getnumberofsongs)
        self.channel.supply(requests.getnumberofalbums, self.getnumberofalbums)
        self.channel.supply(requests.getnumberofartists, self.getnumberofartists)
        self.channel.supply(requests.getnumberoftags, self.getnumberoftags)
        self.channel.supply(requests.getnumberofratings, self.getnumberofratings)

        # and need to be informed about database changes
        self.channel.subscribe(events.dbevent, self.dbevent)

        # finally, we supply some information about the databases and the cache
        self.channel.supply(requests.getsongdbmanagerstats, self.getsongdbmanagerstats)

    def resetafterexception(self):
        # when an exception occurs, we clear the cache
        self.requestcache = {}

    def addsongdb(self, id, config):
        """ add songdb with id defined by config
        return id (or None if player is turned off)
        """
        type = config.type
        if type=="off":
            return None

        if type=="local":
            import songdbs.sqlite
            songdb = songdbs.sqlite.songdb(id, config, self.songdbhub)
        elif type=="remote":
            import songdbs.remote
            songdb = songdbs.remote.songdb(id, config.networklocation, self.songdbhub)

        for postprocessor_name in config.postprocessors:
            try:
                metadata.get_metadata_postprocessor(postprocessor_name)
            except:
                raise RuntimeError("Unkown metadata postprocesor '%s' for database '%r'" % (postprocessor_name, id))

        self.songdbids.append(id)
        songdb.setName("song database thread (id=%s)" % id)
        songdb.start()
        if config.autoregisterer:
            hub.notify(events.autoregistersongs(id))

        return id

    # method decorators for result caching and random song selection

    def cacheresult(requesthandler):
        """ method decorator which caches results of the request """
        def newrequesthandler(self, request):
            log.debug("dbrequest cache: query for request: %r" % request)
            requesthash = hash(request)
            log.debug("dbrequest cache: sucessfully hashed request: %d" % requesthash)
            try:
                # try to get the result from the cache
                result = self.requestcache[requesthash][0]
                # update atime
                self.requestcache[requesthash][2] = time.time()
                self.requestcachehits += 1
                log.debug("dbrequest cache: hit for request: %r" % request)
            except KeyError:
                # make a copy of request for later storage in cache
                requestcopy = copy.copy(request)
                result = requesthandler(self, request)
                resultnoobjects = len(gc.get_referents(result)) + 1
                self.requestcache[requesthash] = [result, requestcopy, time.time(), resultnoobjects]
                self.requestcachemisses += 1
                self.requestcachesize += resultnoobjects
                # remove least recently used items from cache
                if self.requestcachesize > self.requestcachemaxsize:
                    log.debug("dbrequest cache: purging old items")
                    cachebytime = [(item[2], key) for key, item in self.requestcache.items()]
                    cachebytime.sort()
                    for atime, key in cachebytime[-10:]:
                        self.requestcachesize -= self.requestcache[key][3]
                        del self.requestcache[key]
                log.debug("db request cache miss for request: %r (%d requests and %d objects cached)" %
                          (request, len(self.requestcache), self.requestcachesize))
            return result
        return newrequesthandler

    def _genrandomchoice(self, songs):
        """ returns random selection of songs up to the maximal length
        configured. Note that this method changes as a side-effect the
        parameter songs"""

        # consider trivial case separately
        if not songs:
            return []

        # choose item, avoiding duplicates. Stop after a predefined
        # total length (in seconds). Take rating of songs/albums/artists
        # into account
        length = 0
        result = []

        # generate an initial random sample of large enough size samplesize 
        # to choose from
        samplesize = min(100, len(songs))
        sample = random.sample(songs, samplesize)
        currenttime = time.time()

        # relative percentage of songs accepted with a given rating
        ratingdistribution = [5, 10, 20, 30, 35]

        # normalize distribution
        normfactor = float(sum(ratingdistribution))
        ratingdistribution = [x/normfactor for x in ratingdistribution]

        # scale for rating reduction: for playing times longer
        # ago than lastplayedscale seconds, the rating is not
        # influenced.
        lastplayedscale = 60.0 * 60 * 24

        while length < config.general.randominsertlength:
            for song in sample:
                # we have to query the song from our databases 
                # since otherwise this is done automatically leading to
                # a deadlock
                if song.song_metadata is None:
                    song.song_metadata = self.songdbhub.request(requests.getsong_metadata(song.songdbid, song.id))
                    # if the song has been deleted in the meantime, we proceed to the next one
                    if song.song_metadata is None:
                        continue
                if song.rating:
                    rating = song.rating
                else:
                    # punish skipped songs if they have not been rated
                    rating = min(5, max(1, 3 + max(0, 0.5*(song.playcount - song.skipcount))))
                if song.date_lastplayed:
                    # Simple heuristic algorithm to consider song ratings
                    # for random selection. Certainly not optimal!
                    last = max(0, (currenttime-song.date_lastplayed)/60)
                    rating -= 2 * math.exp(-last/lastplayedscale)
                    if rating < 1:
                        rating = 1
                if rating == 5:
                    threshold = ratingdistribution[4]
                else:
                    # get threshold by linear interpolation
                    intpart = int(rating)
                    rest = rating-intpart
                    threshold = (ratingdistribution[intpart-1] +
                                 (ratingdistribution[intpart] - ratingdistribution[intpart-1])*rest)
                if random.random() <= threshold or len(sample) == 1:
                    result.append(song)
                    length += song.length
                    if length >= config.general.randominsertlength or len(result) >= samplesize:
                        return result
            # recreate sample without already chosen songs, if we ran out of songs
            sample = [song for song in sample if song not in result]
        return result


    def selectrandom(requesthandler):
        """ method decorator which returns a random selection of the request result if requested

        Note that the result has to be a list of songs.
        """
        def newrequesthandler(self, request):
            songs = requesthandler(self, request)
            if request.random:
                 return self._genrandomchoice(songs)
            else:
                 return songs
        return newrequesthandler

    def sortresult(requesthandler):
        """ method decorator which sorts the result list if requested """
        def newrequesthandler(self, request):
            result = requesthandler(self, request)
            # XXX turned off
            if request.sort and 0:
                result.sort(request.sort)
            return result
        return newrequesthandler

    # cache update

    def updatecache(self, event):
        """ update/clear requestcache when database event sent """
        if isinstance(event, (events.checkpointdb, events.autoregistersongs)):
            return
        if isinstance(event, events.update_song):
            oldsong_metadata = self.songdbhub.request(requests.getsong_metadata(event.songdbid, event.song.id))
            newsong_metadata = event.song.song_metadata
            # The following is an optimization for an update_song event which occurs rather often
            # Not very pretty, but for the moment enough
            if ( oldsong_metadata.album == newsong_metadata.album and
                 oldsong_metadata.artist == newsong_metadata.artist and
                 oldsong_metadata.tags == newsong_metadata.tags and
                 oldsong_metadata.rating == newsong_metadata.rating ):
                # only the playing information was changed, so we just
                # delete the relevant cache results

                for key, item in self.requestcache.items():
                    if isinstance(item[1], (requests.getsongs)):
                        del self.requestcache[key]
                return
        # otherwise we delete the queries for the correponding database (and all compound queries)
        log.debug("dbrequest cache: emptying cache for database %r" % event.songdbid)
        for key, item in self.requestcache.items():
            songdbid = item[1].songdbid
            if songdbid is None or songdbid == event.songdbid:
                del self.requestcache[key]

    # event handlers

    def quit(self, event):
        service.service.quit(self, event)
        self.songdbhub.notify(events.quit())

    def dbevent(self, event):
        if event.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database event" % event.songdbid)
            return

        # first update result cache (to allow the updatecache method
        # to query the old state of the database)
        self.updatecache(event)
        # and then send the event to the database
        self.songdbhub.notify(event)

    # request handlers

    def dbrequestsingle(self, request):
        if request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
            return

        return self.songdbhub.request(request)

    def dbrequestsongs(self, request):
        # make a copy of the original request, because we will subsequently modify it
        nrequest = copy.copy(request)
        # also reset the sort function as otherwise
        # sending over the network (which requires pickling the
        # request) fails
        # XXX we disable this at the moment
        if request.songdbid is None:
            nrequest.sort = False
            resulthash = {}
            for songdbid in self.songdbids:
                nrequest.songdbid = songdbid
                # Query the songs in the database songdbid via dbrequestsongs to cache
                # the result.
                # Note that in the case of getlastplayedsongs requests, we cheat
                # a little bit, since then the result of the database request is a tuple
                # (playingtime, dbsong).
                for dbsong in self.dbrequestsongs(nrequest):
                    resulthash[dbsong] = songdbid
            return resulthash.values()
        elif request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
            return
        else:
            return self.songdbhub.request(nrequest)
    dbrequestsongs = selectrandom(cacheresult(sortresult(dbrequestsongs)))

    def dbrequestlist(self, request):
        # make a copy of the original request, because we will subsequently modify it
        nrequest = copy.copy(request)

        if request.songdbid is None:
            resulthash = {}
            for songdbid in self.songdbids:
                nrequest.songdbid = songdbid
                for result in self.dbrequestlist(nrequest):
                    resulthash[result] = songdbid
            # sort results
            return resulthash.keys()
        elif request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
        else:
            # use nrequest here instead of request in order to not
            # send sort to database (this fails when
            # using a network channel, since we cannot pickle these
            # objects)
            return self.songdbhub.request(nrequest)
    dbrequestlist = cacheresult(dbrequestlist)

    def getdatabasestats(self, request):
        if request.songdbid is None:
            return "Virtual", ""
        elif request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
        else:
            return self.songdbhub.request(request)

    # requests which return number of items of a certain kind

    def getnumberofsongs(self, request):
        if request.songdbid is not None and request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
        # XXX use filters in sqlite instead
        if request.songdbid is not None and request.filters is None:
            return self.songdbhub.request(request)
        else:
            return len(self.dbrequestsongs(requests.getsongs(songdbid=request.songdbid,
                                                             filters=request.filters)))
    getnumberofsongs = cacheresult(getnumberofsongs)

    def _requestnumbers(self, request, listrequest):
        """ helper method for a request which queries for the number of items.

        If a database is specified, the corresponding database request is
        executed directly. Otherwise, the length of the result of listrequest
        is returned. """
        if request.songdbid is not None and request.songdbid not in self.songdbids:
            log.error("songdbmanager: invalid songdbid '%r' for database request" % request.songdbid)
        elif request.songdbid is None or request.filters:
            return len(self.dbrequestlist(listrequest(songdbid=request.songdbid, filters=request.filters)))
        else:
            return self.songdbhub.request(request)

    def getnumberofalbums(self, request):
        return self._requestnumbers(request, requests.getalbums)
    getnumberofalbums = cacheresult(getnumberofalbums)

    def getnumberoftags(self, request):
        return self._requestnumbers(request, requests.gettags)
    getnumberoftags = cacheresult(getnumberoftags)

    def getnumberofartists(self, request):
        return self._requestnumbers(request, requests.getartists)
    getnumberofartists = cacheresult(getnumberofartists)

    def getnumberofratings(self, request):
        return self._requestnumbers(request, requests.getratings)
    getnumberofratings = cacheresult(getnumberofratings)

    def getsongdbmanagerstats(self, request):
        songdbsstats = []
        for songdbid in self.songdbids:
            songdbsstats.append(self.songdbhub.request(requests.getdatabasestats(songdbid)))
        return songdbmanagerstats(songdbsstats,
                                  self.requestcachesize, self.requestcachemaxsize,
                                  len(self.requestcache),
                                  self.requestcachehits, self.requestcachemisses)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2004, 2005, 2006 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

version = "3.0.3"
copyright = u"(c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Jörg Lehmann <joerg@luga.de>"



# this is required to convert this directory into a package

""" this package contains a collection of services, i.e.
threads processing requests (and acting on events) of
other parts of PyTone. """




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2005 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, threading, traceback
import events, hub, log

class service(threading.Thread):

    def __init__(self, name, daemonize=False, hub=hub._defaulthub):
        threading.Thread.__init__(self)
        # as independent thread, we want our own event and request channel
        # and need at least respond to a quit event
        self.name = name
        self.setName("%s service" % name)
        self.channel = hub.newchannel()
        self.channel.subscribe(events.quit, self.quit)
        self.done = False
        self.setDaemon(daemonize)
        log.debug("started %s service" % self.name)

    def resetafterexception(self):
        """ called after an exception has occured during the event/request handling

        If not reset is possible, an exception can be raised which will lead
        to the termination of the service.
        """
        pass

    def work(self):
        """ do the job """
        self.channel.process(block=True)

    def run(self):
        # main loop of the service
        while not self.done:
            # process events and catch all unhandled exceptions
            try:
                self.work()
            except Exception, e:
                log.debug_traceback()
                self.resetafterexception()

    # event handlers

    def quit(self, event):
        self.done = True



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import SocketServer, socket
import threading, tempfile, time, os.path
import cPickle, cStringIO
import events, hub
import log

# for unpickling
import item, metadata, requests, services, services.player, services.playlist, copy_reg, __builtin__

_EVENT = "EVENT"
_REQUEST = "REQUEST"
_RESULT = "RESULT"
_SUBSCRIBE = "SUBSCRIBE"
_SENDFILE = "SENDFILE"

##############################################################################
# restricted unpickling
##############################################################################


def find_global(module, klass):
    if module in ("events", "requests", "metadata", "copy_reg", "__builtin__"):
        pass
    elif module=="item" and klass=="song":
        pass
    elif module=="dbitem":
        pass
    elif module=="services.player":
        pass
    elif module=="services.playlist":
        pass
    else:
        log.debug("refusing to unpickle %s.%s" % (module, klass))
        raise cPickle.UnpicklingError, \
              "cannot unpickle a %s.%s" % (module, klass)
    log.debug("unpickling %s.%s" % (module, klass))
    return eval("%s.%s" % (module, klass))

def loads(s):
    unpickler = cPickle.Unpickler(cStringIO.StringIO(s))
    unpickler.find_global = find_global
    return unpickler.load()


##############################################################################
# server part
##############################################################################

class servernetworkreceiver(threading.Thread):
    """ helper thread that processes requests coming from clients

    We need this, since there is no select which accepts both a socket and
    a queue.

    """

    def __init__(self, socket, handler):
	self.socket = socket
	self.handler = handler
	self.rfile = self.socket.makefile("r")
	self.done = False
	threading.Thread.__init__(self)
	self.setDaemon(1)

    def _receiveobject(self):
	line = self.rfile.readline()
        if not line: return None, None
	try:
            log.debug("server: request type received")
	    type, bytes = line.split()
	    bytes = int(bytes)
	    if type != _SENDFILE:
 		objstring = self.rfile.read(bytes+2)[:-2]
                log.debug("server: object received")
                obj = loads(objstring)
                log.debug("server receive: type=%s object=%s" % (type, `obj`))
		return (type, obj)
	    else:
		# we handle send file requests separately
		filename = self.rfile.readline()
		tmpfilename = tempfile.mktemp()
		bytes = bytes-len(filename)-2
		tmpfile = open(tmpfilename, "w")
		while bytes>2:
		    rbytes = min(bytes, 4096)
		    tmpfile.write(self.rfile.read(rbytes))
		    bytes -= rbytes
		self.rfile.read(2)
		return (type, tmpfilename)
	except Exception, e:
            log.debug("exception '%s' occured during _receiveobject" % e)
	    return (None, None)

    def run(self):
        # process events, request and subscription requests coming from
        # the client
	while not self.done:
	    type, obj = self._receiveobject()
	    if type == _EVENT:
                log.debug("server: client sends event '%s'" % obj)
		hub.notify(obj, priority=-50)
	    elif type == _REQUEST:
                log.debug("server: requesting %s for client" % `obj`)
		# extract id
		rid, obj = obj
		result = hub.request(obj, priority=-50)
		log.debug("server: got answer %s" % `result`)
		# be careful, handler may not exist anymore?
		try:
		    self.handler._sendobject(_RESULT, (rid, result))
		except:
		    pass
	    elif type == _SUBSCRIBE:
		log.debug("server: client requests subscription for '%s'" % `obj`)
		# be careful, maybe handler does not exists anymore?
		try:
		    self.handler.subscribe(obj)
		except:
		    pass
	    else:
                log.debug("server: servernetworkreceiver exits: type=%s" % type)
		self.done = True
		self.handler.done = True


class handler(SocketServer.StreamRequestHandler, SocketServer.BaseRequestHandler):
    """ handles requests by clients """
    rbufsize = 0

    def _sendobject(self, type, obj):
	# we have to switch to blocking mode for send
	# self.request.setblocking(1)
	objstring = cPickle.dumps(obj, 1)
	self.wfile.write("%s %d\r\n%s\r\n" % (type, len(objstring), objstring))
	self.wfile.flush()
	log.debug("server send: type=%s object=%s" % (type, `obj`))

    def handle(self):
        log.debug("starting handler")
	self.channel = hub.newchannel()
	self.done = False
	self.servernetworkreceiver = servernetworkreceiver(self.request, self)
	self.servernetworkreceiver.start()

        # Process events coming from the rest of the PyTone server.
        # This sends (via eventhandler) subscribed events to the client
	while not self.done:
	    self.channel.process(block=True)
	log.debug("terminating handler")
	self.channel.hub.disconnect(self.channel)

    def subscribe(self, event):
        # clientnetworkreceiver calls this method to subscribe to certain events
	self.channel.subscribe(event, self.eventhandler)

    #
    # event handler
    #

    def eventhandler(self, event):
        # send every subscribed event to client
	log.debug("network event handler called")
	self._sendobject(_EVENT, event)

# boilerplate server code

class tcpserver(threading.Thread):
    allow_reuse_address = 1
    def __init__(self, bind, port):
        self.bind = bind
	self.port = port
	threading.Thread.__init__(self)
	self.setDaemon(1)

    def run(self):
	while 1:
	    try:
		self.tcpserver = SocketServer.ThreadingTCPServer((self.bind, self.port), handler)
		break
	    except:
		log.debug("server thread is waiting for port to become free")
		time.sleep(1)
	self.tcpserver.serve_forever()

class unixserver(threading.Thread):
    def __init__(self, filename):
	self.filename = filename
        try:
            os.unlink(self.filename)
        except OSError, e:
            if e.errno!=2:
                raise
	threading.Thread.__init__(self)
	self.setDaemon(1)

    def run(self):
        self.unixserver = SocketServer.ThreadingUnixStreamServer(self.filename, handler)
	self.unixserver.serve_forever()

##############################################################################
# client part
##############################################################################

class clientnetworkreceiver(threading.Thread):
    """ helper thread that receives from socket and puts result in queue

    We need this, since there is no select which accepts both a socket and
    a queue.

    """

    def __init__(self, socket, queue):
	self.socket = socket
	self.rfile = self.socket.makefile("r")
	self.queue = queue
	self.done = False
	threading.Thread.__init__(self)
	self.setDaemon(1)

    def _receiveobject(self):
	try:
	    line = self.rfile.readline()
	    type, bytes = line.split()
	    bytes = int(bytes)
	    objstring = self.rfile.read(bytes+2)[:-2]
	    log.debug("client receive: %s bytes" % len(objstring))
	    obj = loads(objstring)
	    log.debug("client receive: type=%s object=%s" % (type, repr(obj)))
	    return (type, obj)
	except:
	    return (None, None)

    def run(self):
	while not self.done:
	    self.queue.put((self._receiveobject(), 100))

#
# bidirectional (sending + receiving) client functionality is provided by the clientchannel
# and its subclasses
#

class clientchannel(threading.Thread):
    def __init__(self, networklocation):
        # network location is either a tuple (server adress, port) or a
        # filename pointing to a socket file
        try:
            server, port = networklocation
            family = socket.AF_INET
        except ValueError:
            filename = networklocation
            family = socket.AF_UNIX
            
        self.socket = socket.socket(family, socket.SOCK_STREAM)
        if family == socket.AF_INET:
            self.socket.connect((server, port))
        else:
            self.socket.connect(filename)

	self.subscriptions = []
	self.wfile = self.socket.makefile("wb")
	self.queue = hub.PriorityQueue(-1)
	self.clientnetworkreceiver = clientnetworkreceiver(self.socket, self.queue)
	self.clientnetworkreceiver.start()
	# hash for pending requests
	self.pendingrequests = {}
	self.done = False
	threading.Thread.__init__(self)
	self.setDaemon(1)
        log.debug("Network clientchannel initialized")

    def _sendobject(self, type, obj):
        log.debug("client send: type=%s object=%s" % (type, obj))
        try:
            objstring = cPickle.dumps(obj, cPickle.HIGHEST_PROTOCOL)
        except Exception, e:
            log.debug_traceback()
	self.wfile.write("%s %d\r\n%s\r\n" % (type, len(objstring), objstring))
	self.wfile.flush()

    def sendfile(self, filename):
	basename = os.path.basename(filename)
	file = open(filename, "r")
	f.seek(0, 2)
	filelen = f.tell()
	f.seek(0, 0)
	# length of request
	rlen = len(basename) + 2 + filelen
	self.wfile.write("%s %d\r\n%s\r\n" % (_SENDFILE, rlen, basename))
	while filelen>0:
	    wbytes = min(filelen, 4096)
	    self.wfile.write(file.read(wbytes))
	    filelen -= wbytes
	self.wfile.write("\r\n")
	self.wfile.flush()
	log.debug("client send: type=%s object=file:%s" % (type, filename))

    def subscribe(self, eventtype, handler):
	# Note that the subscription semantics is a little bit different compared
	# with that of hub.py. The clientchannel is a thread of its own, so
	# it calls the playback without being in a process method!
	self._sendobject(_SUBSCRIBE, eventtype)
	self.subscriptions.append((eventtype, handler))

    def run(self):
	while not self.done:
	    item = self.queue.get()
	    if isinstance(item, events.event):
                self._sendobject(_EVENT, item)
	    elif isinstance(item, hub.requestresponse):
		# send request including id
		rid = id(item)
                log.debug("Sending request (id=%d)" % rid)
		self._sendobject(_REQUEST, (rid, item.request))
		self.pendingrequests[rid] = item
	    else: # input from networkreceiver: tuple (type, obj)
		type, obj = item
		if type==_EVENT:
                    log.debug("Received event from networkreceiver")
		    try:
			for subscribedevent, handler in self.subscriptions:
			    if isinstance(obj, subscribedevent):
				handler(obj)
		    except TerminateEventProcessing:
			pass
		elif type==_RESULT:
		    rid, obj = obj
                    log.debug("Received request result (id=%d) from networkreceiver" % rid)
		    item = self.pendingrequests[rid]
		    item.result = obj
		    item.ready.set()
		    del self.pendingrequests[rid]

    def notify(self, item, priority=0):
	""" notify channel of item (event or request) """
	self.queue.put((item, -priority))

    def request(self, request, priority=0):
	""" submit a request (blocking)

	this method submits a request, waits for the result and
	returns it.  Requests with a high priority are treated first.
	"""
	# generate a request response object for the request,
	# send it to hub and wait for result
	rr = hub.requestresponse(request)
	self.notify(rr, priority)
	rr.waitforcompletion()
	return rr.result

    def quit(self):
        while not self.queue.empty():
            time.sleep(0.1)
        self.done = True
        
#
# unidirectional (only sending) client functionality is provided by the sender
# and its subclasses
#

class sender:
    def __init__(self, socket):
        self.socket = socket
	self.wfile = self.socket.makefile("wb")
        log.debug("Network sender initialized")
    
    def sendevent(self, event):
	objstring = cPickle.dumps(event, 1)
	self.wfile.write("%s %d\r\n%s\r\n" % (_EVENT, len(objstring), objstring))
	self.wfile.flush()

    def close(self):
        self.socket.close()


class tcpsender(sender):
    def __init__(self, server, port):
	asocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	asocket.connect((server, port))
        sender.__init__(self, asocket)


class unixsender(sender):
    def __init__(self, filename):
        asocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	asocket.connect(filename)
        sender.__init__(self, asocket)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2005, 2006, 2007 Jörg Lehmann <joerg@luga.de>
#
# Ogg Vorbis interface by Byron Ellacott <bje@apnic.net>.
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os.path, re, struct, string, time
import encoding
import log

# artist name for compilations
VARIOUS = u"___VARIOUS___"

tracknrandtitlere = re.compile("^\[?(\d+)\]? ?[- ] ?(.*)\.(mp3|ogg)$")

##############################################################################
# song metadata class
##############################################################################

class song_metadata:

    url = None
    type = None
    title = None
    album = None
    artist = None
    album_artist = None
    tags = None
    year = None
    comments = []       # list of tuples (language, description, text)
    lyrics = []         # list of tuples (language, description, text)
    bpm = None
    tracknumber = None
    trackcount = None
    disknumber = None
    diskcount = None
    compilation = False
    length = None
    size = None
    bitrate = None
    is_vbr = None
    samplerate = None
    rating = None
    replaygain_track_gain = None
    replaygain_track_peak = None
    replaygain_album_gain = None
    replaygain_album_peak = None
    date_added = None
    date_updated = None
    dates_played = []
    # the following two items are redundant but stored for efficieny reasons
    date_lastplayed = None
    playcount = 0                 # times fully played
    skipcount = 0                 # times skipped

    def __init__(self):
        self.tags = []
        self.date_updated = self.date_added = time.time()

    def __repr__(self):
        return "metadata(%r)" % (self.url)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)

    def update(self, other):
        " merge filesystem metadata from other into self "
        assert other.url == self.url, RuntimeError("song urls changed")
        self.title = other.title
        self.album = other.album
        self.artist = other.artist
        self.album_artist = other.album_artist
        # keep user tags
        usertags = [tag for tag in self.tags if tag[:2] == "U:"]
        self.tags = other.tags + usertags
        self.year = other.year
        self.comments = other.comments
        self.lyrics = other.lyrics
        self.bpm = other.bpm
        self.tracknumber = other.tracknumber
        self.trackcount = other.trackcount
        self.disknumber = other.disknumber
        self.diskcount = other.diskcount
        self.compilation = other.compilation
        self.length = other.length
        self.size = other.size
        self.bitrate = other.bitrate
        self.is_vbr = other.is_vbr
        self.samplerate = other.samplerate
        self.replaygain_track_gain = other.replaygain_track_gain
        self.replaygain_track_peak = other.replaygain_track_peak
        self.replaygain_album_gain = other.replaygain_album_gain
        self.replaygain_album_peak = other.replaygain_album_peak
        self.date_updated = other.date_updated

##############################################################################
# registry for metadata decoders for various fileformats
##############################################################################

# mapping: file type -> (metadata, decoder class, file extension)
_fileformats = {}

def registerfileformat(type, metadataclass, extension):
    _fileformats[type] = (metadataclass, extension)

def getmetadatadecoder(type):
    return _fileformats[type][0]

def getextensions():
    result = []
    for decoder_function, extension in _fileformats.values():
        result.append(extension)
    return result

def gettype(extension):
    for type, extensions in _fileformats.items():
        if extension.lower() in extensions:
            return type
    return None

##############################################################################
# registry for metadata postprocessors for metadata
##############################################################################

_metadata_postprocessors = {}
# mapping: metadata postprocessor name -> metadata postprocessing function

def register_metadata_postprocessor(name, metadata_postprocessor):
    """ register a metadata postprocessor function of the given name

    - The name must not contain any whitespace 
    - metadata_postprocessor has to be a callable accepting exactly one
      parameter which will be an instance of the metadata class. 
    """
    _metadata_postprocessors[name] = metadata_postprocessor

def get_metadata_postprocessor(name):
    return _metadata_postprocessors[name]

##############################################################################
# factory function for song metadata
##############################################################################

def metadata_from_file(relpath, basedir, tracknrandtitlere, postprocessors):
    """ create song metadata from given file with relative (to basedir) path 
    relpath applying the given list of postprocessors"""

    path = os.path.normpath(os.path.join(basedir, relpath))
    if not os.access(path, os.R_OK):
        raise IOError("cannot read song")

    md = song_metadata()
    md.size = os.stat(path).st_size
    md.type = gettype(os.path.splitext(path)[1])

    read_path_metadata(md, relpath, tracknrandtitlere)

    try:
        metadatadecoder = getmetadatadecoder(md.type)
    except:
        raise RuntimeError("Support for %s songs not enabled" % md.type)

    try:
        log.debug("reading metadata for %r" % path)
        metadatadecoder(md, path)
        log.debug("metadata for %r read successfully" % path)
    except:
        log.warning("could not read metadata for %r" % path)
        log.debug_traceback()

    # strip leading and trailing whitespace
    if md.title:
        md.title = md.title.strip()
    if md.artist:
        md.artist = md.artist.strip()
    if md.album:
        md.album = md.album.strip()

    if md.length is None:
        log.warning("could not read length of song %r" % path)
        raise RuntimeError("could not read length of song %r" % path)

    for postprocessor_name in postprocessors:
        try:
            get_metadata_postprocessor(postprocessor_name)(md)
        except:
            log.warning("Postprocessing of song %r metadata with '%r' failed" % (path, postprocessor_name))
            log.debug_traceback()

    # set album_artist if not present
    if md.album_artist is None:
        if md.compilation: 
            md.album_artist = VARIOUS
        else:
            md.album_artist = md.artist

    return md

def read_path_metadata(md, relpath, tracknrandtitlere):
    relpath = os.path.normpath(relpath)

    md.url = u"file://" + encoding.decode_path(relpath)

    # guesses for title and tracknumber using the filename
    match = re.match(tracknrandtitlere, os.path.basename(relpath))
    if match:
        fntracknumber = int(match.group(1))
        fntitle = match.group(2)
    else:
        fntracknumber = None
        fntitle = os.path.basename(relpath)
        if fntitle.lower().endswith(".mp3") or fntitle.lower().endswith(".ogg"):
            fntitle = fntitle[:-4]

    first, second = os.path.split(os.path.dirname(relpath))
    if first and second and not os.path.split(first)[0]:
        fnartist = first
        fnalbum = second
    else:
        fnartist = fnalbum = ""

    # now convert this to unicode strings using the standard filesystem encoding
    fntitle = encoding.decode_path(fntitle)
    fnartist = encoding.decode_path(fnartist)
    fnalbum = encoding.decode_path(fnalbum)

    fntitle = fntitle.replace("_", " ")
    fnalbum = fnalbum.replace("_", " ")
    fnartist = fnartist.replace("_", " ")

    if fntitle:
        md.title = fntitle
    if fnartist:
        md.artist = fnartist
    if fnalbum:
        md.album = fnalbum
    if fntracknumber:
        md.tracknumber = fntracknumber


# accent_trans = string.maketrans('ÁÀÄÂÉÈËÊÍÌÏÎÓÒÖÔÚÙÜÛáàäâéèëêíìïîóòöôúùüû',
#                                'AAAAEEEEIIIIOOOOUUUUaaaaeeeeiiiioooouuuu')

##############################################################################
# ID3 metadata decoder (using mutagen module)
##############################################################################

def _splitnumbertotal(s):
    """ split string into number and total number """
    r = map(int, s.split("/"))
    number = r[0]
    if len(r) == 2:
        count = r[1]
    else:
        count = None
    return number, count

_mutagen_framemapping = { "TIT2": "title",
                          "TALB": "album",
                          "TPE1": "artist" }

def read_mp3_mutagen_metadata(md, path):

    mp3 = mutagen.mp3.MP3(path, ID3=ID3hack)

    # we definitely want the MP3 header data, even if no ID3 tag is present,
    # so extract this info before anything goes wrong
    md.length = mp3.info.length
    md.samplerate = mp3.info.sample_rate
    md.bitrate = mp3.info.bitrate
    md.comments = []
    md.lyrics = []

    if mp3.tags:
        for frame in mp3.tags.values():
            if frame.FrameID == "TCON":
                genre = " ".join(frame.genres)
                if genre:
                    md.tags.append("G:%s" % genre)
            elif frame.FrameID == "RVA2":
                if frame.channel == 1:
                    if frame.desc == "album":
                        basename = "replaygain_album_"
                    else:
                        # for everything else, we assume it's track gain
                        basename = "replaygain_track_"
                    md[basename+"gain"] = frame.gain
                    md[basename+"peak"] = frame.peak
            elif frame.FrameID == "TLEN":
                try:
                    # we overwrite the length which maybe has been defined above
                    md.length = int(+frame/1000)
                except:
                    pass
            elif frame.FrameID == "TRCK":
                md.tracknumber, md.trackcount = _splitnumbertotal(frame.text[0])
            elif frame.FrameID == "TPOS":
                md.disknumber, md.diskcount = _splitnumbertotal(frame.text[0])
            elif frame.FrameID == "TBPM":
                md.bpm = int(+frame)
            #elif frame.FrameID == "TCMP":
            #   self.compilation = True
            elif frame.FrameID == "TDRC":
                try:
                    md.year = int(str(frame.text[0]))
                except:
                    pass
            elif frame.FrameID == "USLT":
                md.lyrics.append((frame.lang, frame.desc, frame.text))
            elif frame.FrameID == "COMM":
                md.comments.append((frame.lang, frame.desc, " / ".join(frame.text)))
            else:
                name = _mutagen_framemapping.get(frame.FrameID, None)
                if name:
                    text = " ".join(map(unicode, frame.text))
                    md[name] = text
    else:
        log.debug("Could not read ID3 tags for song '%r'" % path)


##############################################################################
# ID3 metadata decoder (using eyeD3 module)
##############################################################################

def read_mp3_eyeD3_metadata(md, path):
    mp3file = eyeD3.Mp3AudioFile(path)
    mp3info = mp3file.getTag()

    # we definitely want the length of the MP3 file, even if no ID3 tag is present,
    # so extract this info before anything goes wrong
    md.length = mp3file.getPlayTime()

    md.is_vbr, bitrate = mp3file.getBitRate()
    md.bitrate = bitrate * 1000
    md.samplerate = mp3file.getSampleFreq()
    md.comments = []
    md.lyrics = []

    if mp3info:
        md.title = mp3info.getTitle()
        md.album = mp3info.getAlbum()
        md.artist = mp3info.getArtist()
        try:
            md.year = int(mp3info.getYear())
        except:
            pass
        try:
            genre = mp3info.getGenre()
        except eyeD3.tag.GenreException, e:
            genre = e.msg.split(':')[1].strip()
        if genre:
            md.tags.append("G:%s" % genre)

        md.tracknumber, md.trackcount = mp3info.getTrackNum()
        md.disknumber, md.diskcount = mp3info.getDiscNum()

        # if the playtime is also in the ID3 tag information, we
        # try to read it from there
        if mp3info.frames["TLEN"]:
            length = None
            try:
                length = int(int(mp3info.frames["TLEN"])/1000)
            except:
                # time in seconds (?), possibly with bad decimal separator, e.g "186,333"
                try:
                    length = int(float(mp3info.frames["TLEN"].replace(",", ".")))
                except:
                    pass
            if length:
                md.length = length
        md.lyrics = u"".join(mp3info.getLyrics())
        md.comments = u"".join(map(lambda c: c.render(), mp3info.getComments()))
        md.bpm = mp3info.getBPM()

        for rva2frame in mp3info.frames["RVA2"]:
            # since eyeD3 currently doesn't support RVA2 frames, we have to decode
            # them on our own following mutagen
            desc, rest = rva2frame.data.split("\x00", 1)
            channel = ord(rest[0])
            if channel == 1:
                gain = struct.unpack('>h', rest[1:3])[0]/512.0
                # http://bugs.xmms.org/attachment.cgi?id=113&action=view
                rest = rest[3:]
                peak = 0
                bits = ord(rest[0])
                bytes = min(4, (bits + 7) >> 3)
                shift = ((8 - (bits & 7)) & 7) + (4 - bytes) * 8
                for i in range(1, bytes+1):
                    peak *= 256
                    peak += ord(rest[i])
                peak *= 2**shift
                peak = (float(peak) / (2**31-1))
                if desc == "album":
                    basename = "replaygain_album_"
                else:
                    # for everything else, we assume it's track gain
                    basename = "replaygain_track_"
                md[basename+"gain"] = gain
                md[basename+"peak"] = peak

try:
    import mutagen.mp3
    import mutagen.id3

    # copied from quodlibet
    class ID3hack(mutagen.id3.ID3):
        "Override 'correct' behavior with desired behavior"
        def loaded_frame(self, tag):
            if len(type(tag).__name__) == 3: tag = type(tag).__base__(tag)
            if tag.HashKey in self and tag.FrameID[0] == "T":
                self[tag.HashKey].extend(tag[:])
            else: self[tag.HashKey] = tag

    registerfileformat("mp3", read_mp3_mutagen_metadata, ".mp3")
    log.info("MP3 support enabled: using mutagen module for id3 tag parsing")
except ImportError:
    try:
        import eyeD3
        registerfileformat("mp3", read_mp3_eyeD3_metadata, ".mp3")
        log.info("MP3 support enabled: using eyeD3 module for id3 tag parsing")
    except ImportError:
        log.info("MP3 support disabled: no metadata reader module found")


##############################################################################
# Ogg Vorbis metadata decoder
##############################################################################

def read_vorbis_metadata(md, path):
    vf = ogg.vorbis.VorbisFile(path)

    # We use the information for all streams (not stream 0).
    # XXX Is this correct?
    md.length = int(vf.time_total(-1))
    md.bitrate = vf.bitrate(-1)

    md.samplerate = vf.info().rate
    md.is_vbr = vf.info().bitrate_lower != vf.info().bitrate_upper

    for name, value in vf.comment().as_dict().items():
        value = value[0]
        if name == "TITLE": md.title = value
        if name == "ALBUM": md.album = value
        if name == "ARTIST": md.artist = value
        if name == "DATE": 
            try: md.year = int(value)
            except ValueError: pass
        if name == "GENRE" and value:
            md.tags.append("G:%s" % value)
        if name == "COMMENT":
           md.comments = [value]
        if name == "TRACKNUMBER":
            try: md.tracknumber = int(value)
            except ValueError: pass
        if name == "TRACKCOUNT":
            try: md.trackcount = int(value)
            except ValueError: pass
        if name == "DISCNUMBER":
            try: md.disknumber = int(value)
            except ValueError: pass
        if name == "DISCCOUNT":
            try: md.diskcount = int(value)
            except ValueError: pass
        if name == "BPM":
            try: md.bpm = int(value)
            except ValueError: pass
        if name.startswith("REPLAYGAIN_"):
            # ReplayGain:
            # example format according to vorbisgain documentation
            # REPLAYGAIN_TRACK_GAIN=-7.03 dB
            # REPLAYGAIN_TRACK_PEAK=1.21822226
            # REPLAYGAIN_ALBUM_GAIN=-6.37 dB
            # REPLAYGAIN_ALBUM_PEAK=1.21822226
            try:
                profile, type = name[11:].split("_")
                basename = "replaygain_%s_" % profile.lower()
                if type == "GAIN":
                   md[basename + "gain"] = float(value.split()[0])
                   try:
                       if md[basename + "peak"] is None:
                           md[basename + "peak"] = 1.0
                   except AttributeError:
                       md[basename + "peak"] = 1.0
                if type == "PEAK":
                   md[basename + "peak"] = float(value)
                   try:
                       if md[basename + "gain"] is None:
                           md[basename + "gain"] = 1.0
                   except AttributeError:
                       md[basename + "gain"] = 0.0
            except (ValueError, IndexError):
                pass

        # XXX how is the song lyrics stored?
        # md.lyrics = []

try:
    import ogg.vorbis
    registerfileformat("ogg", read_vorbis_metadata, ".ogg")
    log.info("Ogg Vorbis support enabled")
except ImportError:
    log.info("Ogg Vorbis support disabled: ogg.vorbis module not found")


##############################################################################
# FLAC metadata decoder
##############################################################################

def read_flac_metadata(md, path):
    chain = flac.metadata.Chain()
    chain.read(path)
    it = flac.metadata.Iterator()
    it.init(chain)
    while 1:
        block = it.get_block()
        if block.type == flac.metadata.VORBIS_COMMENT:
            comment = flac.metadata.VorbisComment(block).comments
            id3get = lambda key, default: getattr(comment, key, default)
            self.title = id3get('TITLE', "")
            self.album = id3get('ALBUM', "")
            self.artist = id3get('ARTIST', "")
            self.year = id3get('DATE', "")
            self.genre  = id3get('GENRE', "")
            self.tracknr = id3get('TRACKNUMBER', "")
        elif block.type == flac.metadata.STREAMINFO:
            streaminfo = block.data.stream_info
            self.length = streaminfo.total_samples / streaminfo.sample_rate
        if not it.next():
            break

try:
    import flac.metadata
    registerfileformat("flac", read_flac_metadata, ".flac")
    log.info("FLAC support enabled (VERY EXPERIMENTAL)")
except ImportError:
    log.info("FLAC support disabled: flac module not found")

##############################################################################
# various metadata postprocessors
##############################################################################

def md_pp_capitalize(md):
    def capwords(s):
        # capitalize words also directly after a single punctuation character
        words = []
        for word in s.split():
            if word[0] in string.punctuation:
                words.append(word[0] + word[1:].capitalize())
            else:
                words.append(word.capitalize())
        return " ".join(words)

    if md.title:
        md.title = capwords(md.title)
    if md.artist:
        md.artist = capwords(md.artist)
    if md.album:
        md.album = capwords(md.album)

def md_pp_strip_leading_article(md):
    # strip leading "The " in artist names, often used inconsistently
    if md.artist and md.artist.startswith("The ") and len(md.artist)>4:
        md.artist = md.artist[4:]

def md_pp_remove_accents(md):
    # XXX disabled because I don't know how to get translate working
    # with unicode strings (except for encoding them first)
    md.artist = md.artist.translate(accent_trans)
    md.album = md.album.translate(accent_trans)
    md.title = md.title.translate(accent_trans)

def md_pp_add_decade_tag(md):
    # automatically add decade tag
    if md.year:
        md.tags.append("D:%d" % (10*(md.year//10)))

register_metadata_postprocessor("capitalize", md_pp_capitalize)
register_metadata_postprocessor("strip_leading_article", md_pp_strip_leading_article)
register_metadata_postprocessor("add_decade_tag", md_pp_add_decade_tag)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys, traceback

def formattime(t):
    return "%2d:%02d" % divmod(t, 60)

# Recipe for extended traceback from Python Cookbook

def print_exc_plus():
    tb = sys.exc_info()[2]
    while 1:
        if not tb.tb_next:
            break
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back
    stack.reverse()
    traceback.print_exc(file=sys.stdout)
    print "Locals by frame, innermost last"
    for frame in stack:
        print
        print "Frame %s in %s at line %s" % (frame.f_code.co_name,
                                             frame.f_code.co_filename,
                                             frame.f_lineno)
        for key, value in frame.f_locals.items():
            print "\t%20s = " % key,
            try:
                print value
            except:
                print "<ERROR WHILE PRINTING VALUE>"



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses
import curses.panel
import events, hub
import config
import encoding

class window:
    def __init__(self, screen,
                 h, w, y, x,
                 colors,
                 title=None,
                 border = 15,
                 hasscrollbar=0):
        self.screen = screen
        self.win = curses.newwin(h, w, y, x)
        self.panel = curses.panel.new_panel(self.win)
        self.panel.set_userptr(self)
        curses.panel.update_panels()
        self.colors = colors
        self.border = border
        self.hasscrollbar = hasscrollbar
        self._setdimensions(h, w, y, x)
        self.settitle(title)
        # list of additional border elements due to connections from other windows
        self.borderelements = []

    def _setdimensions(self, h, w, y, x):
        self.h = h
        self.w = w
        self.x = x
        self.y = y

        self.win.bkgdset(0, self.colors.background)

        # coordinates of upper left corner and width and height of inner part of window
        if self.h>=2:
            # we always have a title line
            self.iy = 1
            if self.hasleftborder():
                self.ix = 1
            else:
                self.ix = 0
            if self.hasrightborder() or self.hasscrollbar:
                self.iw = self.w - self.ix - 1
            else:
                self.iw = self.w - self.ix 

            if self.hasbottomborder():
                self.ih = h - self.iy - 1
            else:
                self.ih = h-self.iy
        else:
            # treat very small windows, which do not have a title line, separately
            self.ix = 0
            self.iy = 0
            self.ih = h
            self.iw = w

    # some error protected versions of the standard curses library

    def move(self, *args):
        self.win.move(*args)

    def addch(self, *args):
        try:
            self.win.addch(*args)
        except curses.error:
            pass

    def addstr(self, *args):
        try:
            self.win.addstr(*args)
        except curses.error:
            pass

    def addnstr(self, *args):
        try:
            self.win.addnstr(*args)
        except curses.error:
            pass

    def clrtoeol(self):
        self.win.clrtoeol()

    def clear(self):
        for y in range(self.ih):
            self.addstr(y+1, self.ix, " "*self.iw)

    def hline(self, x, y, c, n, attr):
        try:
            self.win.hline(x, y, c, n, attr)
        except TypeError:
            # workaround for Python 2.1.x
            for y in range(y, y+n):
                self.addch(x, y, c, attr)

    def vline(self, x, y, c, n, attr):
        try:
            self.win.vline(x, y, c, n, attr)
        except TypeError:
            # workaround for Python 2.1.x
            for x in range(x, x+n):
                self.addch(x, y, c, attr)

    def hastopborder(self):
        return self.border & config.BORDER_TOP

    def hasbottomborder(self):
        return self.border & config.BORDER_BOTTOM

    def hasleftborder(self):
        return self.border & config.BORDER_LEFT

    def hasrightborder(self):
        return self.border & config.BORDER_RIGHT

    def getborderends(self):
        """ return open ends of border as list of tuples (y, x, direction)
        where direction is one of "left", "right" , "up" or "down"
        """

        borderends = []
        if self.hastopborder() and not self.hasleftborder():
            borderends.append((self.y, self.x-1, "left"))
        if self.hastopborder() and not self.hasrightborder():
            borderends.append((self.y, self.x+self.w+1, "right"))

        if self.hasbottomborder() and not self.hasleftborder():
            borderends.append((self.y+self.h-1, self.x-1, "left"))
        if self.hasbottomborder() and not self.hasrightborder():
            borderends.append((self.y+self.h-1, self.x+self.w+1, "right"))

        if self.hasleftborder() and not self.hastopborder():
            borderends.append((self.y-1, self.x, "up"))
        if self.hasrightborder() and not self.hastopborder():
            borderends.append((self.y-1, self.x+self.w-1, "up"))

        if self.hasleftborder() and not self.hasbottomborder():
            borderends.append((self.y+self.h+1, self.x, "down"))
        if self.hasrightborder() and not self.hasbottomborder():
            borderends.append((self.y+self.h+1, self.x+self.w-1, "down"))

        return borderends

    def connectborderends(self, borderends):
        """ update list of border connections from other windows """
        self.borderelements = []
        for y, x, d in borderends:
            if self.win.enclose(y, x):
                if d == "right" and self.hasleftborder():
                    dy = y - self.y
                    if dy == 0 and self.hastopborder():
                        cel = curses.ACS_TTEE
                    elif dy == self.h-1 and self.hasbottomborder():
                        cel = curses.ACS_BTEE
                    else:
                        cel = curses.ACS_LTEE
                    self.borderelements.append((dy, 0, cel))
                elif d == "left" and self.hasrightborder():
                    dy = y - self.y
                    if dy == 0 and self.hastopborder():
                        cel = curses.ACS_TTEE
                    elif dy == self.h-1 and self.hasbottomborder():
                        cel = curses.ACS_BTEE
                    else:
                        cel = curses.ACS_LTEE
                    self.borderelements.append((dy, self.w-1, cel))
                elif d == "up" and self.hasbottomborder():
                    dx = x - self.x
                    if dx == 0 and self.hasleftborder():
                        cel = curses.ACS_LTEE
                    elif dx == self.w-1 and self.hasrightborder():
                        cel = curses.ACS_RTEE
                    else:
                        cel = curses.ACS_TTEE
                    self.borderelements.append((self.h, dx, cel))
                elif d == "down" and self.hastopborder():
                    dx = x - self.x
                    if dx == 0 and self.hasleftborder():
                        cel = curses.ACS_LTEE
                    elif dx == self.w-1 and self.hasrightborder():
                        cel = curses.ACS_RTEE
                    else:
                        cel = curses.ACS_BTEE
                    self.borderelements.append((0, dx, cel))

    def resize(self, h, w, y, x):
        """ resize window """
        try:
            self.win.resize(h, w)
            self.win.mvwin(y, x)
        except curses.error:
            pass
        self._setdimensions(h, w, y, x)

    def settitle(self, title):
        if title is not None:
            self.title = title
        else:
            self.title = None

    def scrollbardimensions(self, top, total):
        if total>0:
            totalheight = self.ih-4
            scrollbarbegin = 3 + totalheight*top/total
            scrollbarheight = totalheight*self.ih/total
            scrollbarheight = min(max(scrollbarheight, 1), totalheight-scrollbarbegin+3)

            return scrollbarbegin, scrollbarheight
        else:
            return 0, 0

    def drawscrollbar(self, top, total):
        if self.hasscrollbar:
            if not self.hasrightborder() and self.ih>2:
                self.vline(1, self.iw+self.ix, " ", self.ih, self.colors.background)
            if total>self.ih:
                xpos = self.iw+self.ix
                if top!=0:
                    self.addch(2, xpos, curses.ACS_UARROW, self.colors.scrollbararrow)
                if top+self.ih<total:
                    # self.addch(self.ih-1, xpos, curses.ACS_DARROW, self.colors.scrollbararrow)
                    self.addch(self.ih-1, xpos, "v", self.colors.scrollbararrow)
                scrollbarbegin, scrollbarheight = self.scrollbardimensions(top, total)
                self.vline(3, xpos, curses.ACS_CKBOARD, self.ih-4, self.colors.scrollbar)
                self.vline(scrollbarbegin, xpos, curses.ACS_CKBOARD, scrollbarheight, self.colors.scrollbarhigh)

    def bottom(self):
        """ bring window to bottom"""
        try:
            self.panel.bottom()
        except:
            pass
        else:
            hub.notify(events.focuschanged())

    def top(self):
        """ bring window to top"""
        try:
            self.panel.top()
            # The following call fixes the redrawing problem when switching between
            # the filelist and the playlist window reported by Dag Wieers.
            curses.panel.update_panels()
        except:
            pass
        else:
            hub.notify(events.focuschanged())

    def hide(self):
        """ hide window """
        try:
            self.panel.hide()
        except:
            pass
        else:
            hub.notify(events.focuschanged())

    def hasfocus(self):
        """ is window on top and has thus the current focus? """
        return self.panel is curses.panel.top_panel()

    def enclose(self, y, x):
        "check whether y, x belongs to window (and not to other windows lying above)"

        # walk through panel stack
        p = curses.panel.top_panel()
        while p:
            if p is self.panel:
                return self.win.enclose(y, x)
            elif p.window().enclose(y, x):
                return 0
            p = p.below()
        return 0

    def stdscrtowin(self, y, x):
        begy, begx = self.win.getbegyx()
        return y-begy, x-begx

    def update(self):
        if self.title and self.h>=2:
            if self.hasfocus():
                topborder = 0
                # uncomment this to get the "thick" border of the old times
                # topborder = ord("=")
                attr = self.colors.activeborder
                try:
                    titleattr = self.colors.activetitle
                except AttributeError:
                    titleattr = self.colors.title
            else:
                topborder = 0
                attr = self.colors.border
                titleattr = self.colors.title

            # draw configured borders

            if self.hastopborder():
                self.hline(0, self.ix, curses.ACS_HLINE, self.iw, attr)
                # self.win.border(0, 0, topborder)
                t = encoding.encode(self.title)[:self.w-4]
                pos = (self.w-4-len(t))/2 
                self.addstr(0, pos, "[ %s ]" % t, titleattr)
                if self.hasleftborder():
                    self.addch(0, 0, curses.ACS_ULCORNER, attr)
                if self.hasrightborder():
                    self.addch(0, self.ix+self.iw, curses.ACS_URCORNER, attr)
                else:
                    self.addch(0, self.ix+self.iw, curses.ACS_HLINE, attr)
            else:
                t = self.title[:self.w]
                self.addstr(0, 0, t.center(self.w), titleattr)

            if self.hasbottomborder():
                self.hline(self.iy+self.ih, self.ix, curses.ACS_HLINE, self.iw, attr)
                if self.hasleftborder():
                    self.addch(self.iy+self.ih, 0, curses.ACS_LLCORNER, attr)
                if self.hasrightborder():
                    self.addch(self.iy+self.ih, self.ix+self.iw, curses.ACS_LRCORNER, attr)
                else:
                    self.addch(self.iy+self.ih, self.ix+self.iw, curses.ACS_HLINE, attr)

            if self.hasleftborder():
                if self.hastopborder():
                    self.vline(self.iy, 0, curses.ACS_VLINE, self.ih, attr)
                else:
                    self.vline(0, 0, curses.ACS_VLINE, self.ih+1, attr)

            if self.hasrightborder():
                if self.hastopborder():
                    self.vline(self.iy, self.ix+self.iw, curses.ACS_VLINE, self.ih, attr)
                else:
                    self.vline(0, self.ix+self.iw, curses.ACS_VLINE, self.ih+1, attr)

            # draw additional border elements
            for y, x, c in self.borderelements:
                self.addch(y, x, c, attr)

    # event handler

    def hidewindow(self, event):
        # subclasses of window explicitely have to subscribe to this event,
        # if they need to
        if event.window==self:
            self.hide()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2005 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import encoding
import messagewin
import hub, requests

class statswin(messagewin.messagewin):

    def __init__(self, screen, maxh, maxw, channel, numberofsongdbs):
        # column number of message string
        messagewin.messagewin.__init__(self, screen, maxh, maxw, channel,
                                       config.colors.statswindow,
                                       _("PyTone Statistics"), [],
                                       config.statswindow.autoclosetime)
        self.numberofsongdbs = numberofsongdbs

    def _outputlen(self, iw):
        """number of lines in window with inner widht iw"""
        result = self.numberofsongdbs*4 + 3
        return result

    def showitems(self):
        lines = []
        stats = hub.request(requests.getsongdbmanagerstats())
        indent = " "*3
        for songdbstats in stats.songdbsstats:
            dbidstring = _("Database %s") % songdbstats.id + ":"
            dbstatstring = _("%d songs, %d albums, %d artists, %d tags") % (songdbstats.numberofsongs,
									    songdbstats.numberofalbums,
									    songdbstats.numberofartists,
									    songdbstats.numberoftags)

            lines.append((dbidstring, dbstatstring))
            if songdbstats.type == "local":
                dbtypestring = _("local database (db file: %s)") % (songdbstats.dbfile)
            else:
                dbtypestring = _("remote database (server: %s)") % (songdbstats.location)
            lines.append((indent + _("type") + ":", dbtypestring))
            lines.append((indent + _("base directory") + ":", songdbstats.basedir)) 
            dbcachesizestring =  "%dkB" % songdbstats.cachesize
            lines.append((indent + _("cache size") + ":", dbcachesizestring))

        lines.append(("", ""))
        cachestatsstring = _("%d requests, %d / %d objects") % (stats.requestcacherequests, stats.requestcachesize,
                                                                stats.requestcachemaxsize)
        if stats.requestcachemaxsize != 0:
            cachestatsstring = cachestatsstring + " (%d%%)" % (100*stats.requestcachesize//stats.requestcachemaxsize)
        lines.append((_("Request cache size") + ":", cachestatsstring))
        totalrequests = stats.requestcachehits + stats.requestcachemisses
        if totalrequests != 0:
            percentstring = " (%d%%)" % (100*stats.requestcachehits//totalrequests)
        else:
            percentstring = ""
        lines.append((_("Request cache stats") + ":",
                      (_("%d hits / %d requests") % (stats.requestcachehits, totalrequests)) + percentstring))

        wc1 = max([len(lc) for lc, rc in lines]) + 1
        if wc1 > 0.6*self.iw:
            wc1 = int(0.6*self.iw)
        wc2 = self.iw - wc1
        y = self.iy
        for lc, rc in lines:
            self.move(y, self.ix)
            self.addstr(encoding.encode(lc).ljust(wc1)[:wc1], self.colors.description)
            self.addstr(encoding.encode(rc).ljust(wc2)[:wc2], self.colors.content)
            y += 1



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import os.path
import fcntl
import re
import string
import time

import log
import services.playlist
import hub,requests

from services.player import genericplayer

def makeNonBlocking(fd):
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

startpattern = re.compile(r"^@R MPG123 *[-0-9a-zA-Z\s_]*$")

class player(genericplayer):
    def __init__(self, id, playlistid, autoplay, cmdline):
        self.cmdline = cmdline
        self.initmpg123()
        genericplayer.__init__(self, id, playlistid, autoplay)

    def initmpg123(self):
        """start new mpg123 process"""
        self.pstdin, self.pstdout = os.popen4(self.cmdline + " -R -")
        startline = self.pstdout.readline()
        if not re.match(startpattern, startline):
            raise RuntimeError("cannot initialize player")
        makeNonBlocking(self.pstdout.fileno())

    def closempg123(self):
        """terminate running mpg123 process"""
        if self.pstdin:
            self.pstdin.close()
        if self.pstdout:
            self.pstdout.close()

    def sendmpg123(self, command):
        """send command to mpg123 process"""
        try:
            self.pstdin.write("%s\n" % command)
            self.pstdin.flush()
        except IOError, error:
            # broken pipe => restart player
            if error[0]==32:
                self.closempg123()
                self.initmpg123()
                self.playbackinfo.stopped()
                self.pstdin.write("%s\n" % command)
            else:
                raise

    def receivempg123(self):
        """receive command from mpg123 process"""
        try:
            return self.pstdout.readline()
        except (ValueError, IOError):
            return ""
        

    def play(self):
        """play songs"""
        
        r = self.receivempg123()

        if r=="":
            time.sleep(0.1)
            # we just want to tease mpg123 a bit to check if it is still
            # alive
            self.sendmpg123("")
        elif r.startswith("@F"): 
            pframes, lframes, ptime, ltime = string.split(r[3:])[:4]
            ptime = int(float(ptime))
            self.playbackinfo.updatetime(ptime)
        elif r.startswith("@P"):
            if self.playbackinfo.isplaying() and r[3]=="0":
                self.playbackinfo.stopped()
                self.requestnextsong()
        elif r.startswith("@S"): 
            ( version, layer, samplerate, mode, modeextension,
              bytesperframe, channels, copyrighted, crcprotected, emphasis, bitrate ) = string.split(r[3:])[:11]
            self.framespersecond = 1000.0 / 8 * int(bitrate) / int(bytesperframe)
                    
    def _playsong(self, playlistitemorsong, manual):
        """play event.song next"""
        path = None
        if isinstance( playlistitemorsong, services.playlist.playlistitem ):
            url = playlistitemorsong.song.url
            if url.startswith("file://"):
                dbstats = hub.request(requests.getdatabasestats(playlistitemorsong.song.songdbid))
                path = os.path.join(dbstats.basedir, url[7:])
            else:
                path = url
        else:
            log.warning("mpg123 player: song %s not a playlistitem. Not added!" % repr( playlistitemorsong) )
            return
        self.sendmpg123("L %s" % path)
        self.framespersecond = None
        self.playbackinfo.updatesong(song)

    def _playerunpause(self):
        """unpause playing"""
        self.sendmpg123("P")
        # delete messages coming from mpg123
        time.sleep(0.1)
        while self.receivempg123()!="":
            pass
        self.playbackinfo.playing()

    def _playerpause(self):
        """pause playing"""
        self.sendmpg123("P")
        # delete messages coming from mpg123
        time.sleep(0.1)
        while self.receivempg123()!="":
            pass
        self.playbackinfo.paused()

    def _playerseekrelative(self, seconds):
        if self.framespersecond:
            self.sendmpg123("J %+d" % int(seconds * self.framespersecond))

    def _playerstop(self):
        """stop playing"""
        self.sendmpg123("S")
        # delete messages coming from mpg123
        time.sleep(0.1)
        while self.receivempg123()!="":
            pass
        self.playbackinfo.stopped()

    def _playerquit(self):
        self.sendmpg123("Q")
        self.closempg123()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import help
import events, hub
import messagewin
import statusbar
import encoding

def getitems(section):
    items = []
    if section:
        for function in config.keybindings[section].asdict().keys():
            keys = list(config.keybindings[section][function])
            keys.sort()
            keynames = [help.getkeyname(key) for key in keys]
            descr = help.descriptions[section][function][1]
            items += [(keynames, descr)]

    return items


class helpwin(messagewin.messagewin):

    def __init__(self, screen, maxh, maxw, channel):
        messagewin.messagewin.__init__(self, screen, maxh, maxw, channel,
                                       config.colors.helpwindow,
                                       _("PyTone Help"), [],
                                       config.helpwindow.autoclosetime)

        sbar = statusbar.generatedescription("general", "showhelp")
        hub.notify(events.statusbar_update(2, sbar))

    def showitems(self):
        y = self.iy
        for item in self.items[self.first:]:
            self.addstr(y, 1, " "*self.iw, self.colors.background)
            self.move(y, 1)
            for keyname in item[0][:-1]:
                self.addstr(encoding.encode(keyname), self.colors.key)
                self.addstr("/", self.colors.description)
            self.addstr(encoding.encode(item[0][-1]), self.colors.key)
            self.addnstr(y, 35, encoding.encode(item[1]), self.iw-35)
            y += 1
            if y>=self.iy+self.ih:
                break

    def showhelp(self, context):
        self.items = getitems("general")+getitems(context)
        self.items.sort()
        self.show()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import time
import xmms.control

from services.player import genericplayer

# Note: the implementation xmms player is a bit hackish,
# so don't look too close. At many places, sleeps have
# been interted to get it working, so as I said...

class player(genericplayer):
    def __init__(self, id, playlistid, autoplay, session=0, noqueue=0):
        self.session = session
        self.noqueue = noqueue

        # a mapping path -> song
        self.songs = {} 
        self.initxmms()
        
        genericplayer.__init__(self, id, playlistid, autoplay)

    def initxmms(self):
        if not xmms.control.is_running(self.session):
            abs_prog_name = xmms.control._find_and_check_executable("xmms")
            if not abs_prog_name:
                raise xmms.control.ExecutableNotFound("can't find XMMS executable")
            os.system(abs_prog_name + " >/dev/null 2>/dev/null &")
            while not xmms.control.is_running(self.session):
                time.sleep(0.2)

        xmms.control.playlist_clear(self.session)
        xmms.control.main_win_toggle(0, self.session)
        xmms.control.pl_win_toggle(0, self.session)
        xmms.control.eq_win_toggle(0, self.session)

    # event handler

    def play(self):
        if xmms.control.is_playing(self.session):
            pos = xmms.control.get_playlist_pos(self.session)
            # title = xmms.control.get_playlist_title(pos, self.session)
            # ttime = xmms.control.get_playlist_time(pos, self.session)
            # ptime = xmms.control.get_output_time(self.session)
            # self.playbackinfo = (title, int(ptime/1000), int(ttime/1000))
            path = xmms.control.get_playlist_file(pos, self.session)
            song = self.songs[path]
            ptime = xmms.control.get_output_time(self.session)/1000
            self.playbackinfo.updatesong(song)
            self.playbackinfo.updatetime(ptime)

            # fill up xmms playlist if necessary
            if song.length-ptime<20 and xmms.control.get_playlist_length()<2: 
                self.requestnextsong()
            if pos>0:
                path = xmms.control.get_playlist_file(0, self.session)
                xmms.control.playlist_delete(0, self.session)
                del self.songs[path]
        else:
            self.playbackinfo.updatesong(None)

        time.sleep(0.1)

    def _playsong(self, song, manual):
        if self.noqueue:
            xmms.control.playlist_clear(self.session)
            self.songs = {}
            
        self.songs[song.path] = song
        xmms.control.playlist_add((song.path,), self.session)

        if not xmms.control.is_playing():
            xmms.control.play(self.session)
            # wait a little, so that xmms can start playing
            # and we don't request another song...
            time.sleep(0.5)

    def _playerstart(self):
        # before we start playing, we clear the playlist
        xmms.control.playlist_clear(self.session)
        self.songs = {}

    def _playerpause(self):
        # before we start playing, we clear the playlist
        xmms.control.pause(self.session)

    def _playerunpause(self):
        # before we start playing, we clear the playlist
        xmms.control.play(self.session)

    def _playerseekrelative(self, seconds):
        ptime = xmms.control.get_output_time(self.session)
        pos = xmms.control.get_playlist_pos(self.session)
        ttime = xmms.control.get_playlist_time(pos, self.session)
        time = min(max(ptime + int(seconds * 1000), 0), ttime)
        xmms.control.jump_to_time(time, self.session)
    
    def _playerstop(self):
        xmms.control.playlist_clear(self.session)
        self.songs = {}

    def playernext(self, event):
        if event.playerid==self.id:
            if xmms.control.is_playing(self.session):
                self.requestnextsong()
                # wait a little for the other threads, uuh...
                time.sleep(1)
                self.channel.process()
                time.sleep(0.1)
            xmms.control.playlist_next(self.session)

    def _playerquit(self):
        xmms.control.quit(self.session)




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import log
import messagewin
import time

class logwin(messagewin.messagewin):

    def __init__(self, screen, maxh, maxw, channel):
        # column number of message string
        self.mc = 12
        messagewin.messagewin.__init__(self, screen, maxh, maxw, channel,
                                       config.colors.logwindow,
                                       _("PyTone Messages"),
                                       log.items,
                                       config.logwindow.autoclosetime)

    def _outputlen(self, iw):
        """number of lines in window with inner widht iw"""
        result = 0
        for item in self.items:
            result += len(item[2])/(iw-self.mc+2)+1
        return result

    def showitems(self):
        y = self.iy
        for item in log.items[self.first:]:
            self.addstr(y, 1, " "*self.iw, self.colors.background)
            self.addstr(y, 1, log._desc[item[0]][0].upper(), self.colors.time)
            self.addstr(y, 3, time.strftime("%H:%M:%S", time.localtime(item[1])), self.colors.time)

            if item[0] == log._DEBUG:
                color = self.colors.debug
            elif item[0] == log._INFO:
                color = self.colors.info
            elif item[0] == log._WARNING:
                color = self.colors.warning
            else:
                color = self.colors.error
            # width of message column
            mw = self.iw-self.mc+1
            if len(item[2])<=mw:
                self.addstr(y, self.mc, item[2], color)
            else:
                words = item[2].split()
                s = words.pop(0)
                while words and len(s)+len(words[0])<mw:
                    s += " %s" % words.pop(0)
                self.addstr(y, self.mc, s, color)
                y += 1
                if y>=self.iy+self.ih:
                    break
                self.addstr(y, 1, " "*self.iw, self.colors.background)
                s=" ".join(words)
                self.addnstr(y, self.mc, s, mw, color)

            y += 1
            if y>=self.iy+self.ih:
                break



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2005 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import item
import services.playlist
import events, hub
import window
import messagewin
import encoding

# marker class
class _selection:
    pass
selection = _selection()

class iteminfowin(window.window):
    def __init__(self, screen, layout, channel, playerids, player):
        # player for pre-listening
        self.player = player      
        # list of players for which information can be displayed
        self.playerids = [playerid for playerid in playerids if playerid is not None]

        # hash for items (for each view mode)
        self.items = {}
        self.items[selection] = None
        for playerid in self.playerids:
            self.items[playerid] = None

        # currently active view mode
        self.activeview = selection
        self.keybindings = config.keybindings.general

        h, w, y, x, border = layout
        window.window.__init__(self, screen, h, w, y, x,
                               config.colors.iteminfowindow,
                               _("MP3 Info"), border)

        channel.subscribe(events.selectionchanged, self.selectionchanged)
        channel.subscribe(events.songchanged, self.songchanged)
        channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
        channel.subscribe(events.keypressed, self.keypressed)

    def resize(self, layout):
        h, w, y, x, self.border = layout
        window.window.resize(self, h, w, y, x)

    def update(self):
        # update window title
        aitem = self.items[self.activeview]
        title = _("No song")
        if isinstance(aitem, (item.song, services.playlist.playlistitem)):
            if isinstance(aitem, item.song):
                atype = aitem.type
            else:
                atype = aitem.song.type
            if atype == "mp3":
                title = _("MP3 Info")
            elif atype == "ogg":
                title = _("Ogg Info")
            else:
                title = _("Song Info")
        elif isinstance(aitem, item.diritem):
            title = _("Directory Info")
        if self.activeview != selection:
            title = title + " " + _("[Player: %s]") % self.activeview
        self.settitle(title)

        window.window.update(self)

        # get lines to display
        empty= [["", "", "", ""]]
        if aitem is not None:
            info = aitem.getinfo()
        else:
            info = []
        l = info + empty*(4-len(info))

        colsep = self.iw > 45

        # calculate width of columns
        wc1 = max( len(l[0][0]), len(l[1][0]), len(l[2][0]), len(l[3][0])) + colsep
        wc3 = max( len(l[0][2]), len(l[1][2]), len(l[2][2])) + colsep
        wc4 = 5
        wc4 = max( len(l[0][3]), len(l[1][3]), len(l[2][3]))
        wc2 = self.iw-wc1-wc3-wc4-1

        for lno in range(4):
            self.move(1+lno, self.ix)
            l0 = encoding.encode(l[lno][0])
            l1 = encoding.encode(l[lno][1])
            self.addstr(l0.ljust(wc1)[:wc1], self.colors.description)
            self.addstr(l1.ljust(wc2)[:wc2], self.colors.content)
            self.addch(" ")
            if lno != 3 or isinstance(aitem, item.diritem):
                l2 = encoding.encode(l[lno][2])
                l3 = encoding.encode(l[lno][3])
                self.addstr(l2.ljust(wc3)[:wc3], self.colors.description)
                self.addstr(l3.ljust(wc4)[:wc4], self.colors.content)
            else:
                l2 = encoding.encode(l[3][-2])
                l3 = encoding.encode(l[3][-1])
                # special handling of last line for songs
                wc3 = max(len(l2), 5) + colsep
                wc4 = max(len(l3), 5)

                self.move(1+lno, self.iw-wc3-wc4-1-self.ix)
                self.addch(" ")
                self.addstr(l2.ljust(wc3)[:wc3], self.colors.description)
                self.addstr(l3.ljust(wc4)[:wc4], self.colors.content)

    # event handler

    def selectionchanged(self, event):
        if self.player and self.items[selection] != event.item:
            if isinstance(event.item, item.song):
                hub.notify(events.playerplaysong(self.player, event.item))
            elif isinstance(event.item, services.playlist.playlistitem):
                hub.notify(events.playerplaysong(self.player, event.item.song))
        self.items[selection] = event.item
        self.update()

    def songchanged(self, event):
        # here we assume that a possible change actually has also affected our
        # item, if not we're missing it
        self.update()

    def playbackinfochanged(self, event):
        playerid = event.playbackinfo.playerid
        if  playerid in self.playerids:
            if event.playbackinfo.song != self.items[playerid]:
                self.items[playerid] = event.playbackinfo.song
                if self.activeview == playerid:
                    self.update()

    def keypressed(self, event):
        key = event.key
        if key in self.keybindings["toggleiteminfowindow"]:
            if self.activeview == selection:
                self.activeview = self.playerids[0]
            else:
                i = self.playerids.index(self.activeview)
                if i < len(self.playerids)-1:
                    self.activeview = self.playerids[i+1]
                else:
                    self.activeview = selection
        else:
            return
        self.update()
        raise hub.TerminateEventProcessing


class iteminfowinlong(messagewin.messagewin):

    def __init__(self, screen, maxh, maxw, channel):
        messagewin.messagewin.__init__(self, screen, maxh, maxw, channel,
                                       config.colors.iteminfolongwindow,
                                       _("Item info"), [],
                                       config.iteminfolongwindow.autoclosetime)

        self.item = None

        channel.subscribe(events.selectionchanged, self.selectionchanged)

    def _outputlen(self, width):
        return 16

    def showitems(self):
        # get lines to display
        empty= [["", "", "", ""]]
        if self.item:
            info = self.item.getinfolong()
        else:
            info = []
        l = info + empty*(4-len(info))

        colsep = self.iw > 45

        # calculate width of columns
        wc1 = 0
        wc3 = 0
        wc4 = 0
        for line in info:
            line = map(encoding.encode, line)
            wc1 = max(wc1, len(line[0]))
            wc3 = max(wc3, len(line[2]))
            wc4 = max(wc3, len(line[3]))
        wc1 += colsep
        wc3 += colsep
        wc2 = self.iw-wc1-wc3-wc4-1
        self.clear()
        for lno in range(len(info)):
            line = map(encoding.encode, l[lno])
            self.move(self.iy+lno, self.ix)
            self.addstr(line[0].ljust(wc1)[:wc1], self.colors.description)
            self.addstr(line[1].ljust(wc2)[:wc2], self.colors.content)
            self.addch(" ")
            if lno!=self.ih:
                self.addstr(line[2].ljust(wc3)[:wc3], self.colors.description)
                self.addstr(line[3].ljust(wc4)[:wc4], self.colors.content)
            else:
                # special handling of last line
                wc3 = max(len(line[-2]), 5) + colsep
                wc4 = max(len(line[-1]), 5)

                self.move(1+lno, self.iw-wc3-wc4-1-self.ix)
                self.addch(" ")
                self.addstr(line[-2].ljust(wc3)[:wc3], self.colors.description)
                self.addstr(line[-1].ljust(wc4)[:wc4], self.colors.content)

    def selectionchanged(self, event):
        self.item = event.item



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import exceptions

class pytoneerror(exceptions.Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "PyTone error: %s" % `self.value`

class configurationerror(pytoneerror):
    def __str__(self):
        return "PyTone configuration error: %s" % `self.value`

class databaseerror(pytoneerror):
    def __str__(self):
        return "PyTone database error: %s" % `self.value`
            
class playererror(pytoneerror):
    def __str__(self):
        return "PyTone player error: %s" % `self.value`



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import copy, time

import events, hub, requests
import log
import services.playlist
import service

def initplayer(id, config):
    """ initialize player with id defined by config
    return id (or None if player is turned off)
    """

    # only the first player has a playlist
    if id == "main":
        playlistid = "main"
    else:
        playlistid = None

    type = config.type
    if type=="off":
        return None
    elif type=="internal":
        import players.internal
        driver = config.driver
        if driver in ("alsa09", "alsa"):
            aooptions = {"dev": config.device}
        elif driver=="oss":
            aooptions = {"dsp": config.device}
        elif driver=="sun":
            aooptions = {"dev": config.device}
        else:
            aooptions = {}
        # add options given by user in config file
        for aooption in config.aooptions.split():
            key, value = aooption.split("=")
            aooptions[key] = value
        try:
            p = players.internal.player(id,
                                        playlistid,
                                        autoplay=config.autoplay,
                                        aodevice=driver,
                                        aooptions=aooptions,
                                        bufsize=config.bufsize,
                                        crossfading=config.crossfading,
                                        crossfadingstart=config.crossfadingstart,
                                        crossfadingduration=config.crossfadingduration,
                                        )
        except:
            log.debug_traceback()
            raise RuntimeError("Cannot initialize %s player: type=internal, device=%s" % (id, config.device))
    elif type=="xmms":
        import players.xmmsplayer
        try:
            p = players.xmmsplayer.player(id,
                                          playlistid,
                                          autoplay=config.autoplay,
                                          session=config.session,
                                          noqueue=config.noqueue)
        except:
            log.debug_traceback()
            raise RuntimeError("Cannot initialize %s player: type=xmms, session=%d" % (id, config.session))
    elif type=="mpg123":
        import players.mpg123
        try:
            p = players.mpg123.player(id,
                                      playlistid,
                                      autoplay=config.autoplay,
                                      cmdline=config.cmdline)

        except:
            log.debug_traceback()
            raise RuntimeError("Cannot initialize %s player: type=mpg123, cmdline=%s" % (id, config.cmdline))
    elif type=="remote":
        import players.remote
        try:
            p = players.remote.player(id, playlistid, config.networklocation)
        except:
            log.debug_traceback()
            raise RuntimeError("Cannot initialize %s player: type=remote, location=%s" % (id, config.networklocation))

    p.setName("player thread (id=%s)" % id)

    if type != "remote" and id == "main":
        services.playlist.initplaylist(id, id, id)

    # start player only after the playlist service has been started since
    # playlist requests may already be issued during the player startup
    p.start()

    return id

# player states

STOP     = 0
PAUSE    = 1
PLAY     = 2

class playbackinfo:
    
    """ class for storage of playback information

    This class serves as a means of communication between the
    actual players and the player control logic 

    """

    def __init__(self, playerid, send_played_skipped_events):
        """ 

        playerid:     player which this playbackinfo instance refers to
        state:        player state (STOP, PAUSE, PLAY)
        song:         song currently played (or None, if player is not playing)
        time:         position in seconds in the song
        crossfade:    crossfade in progress
        send_played_skipped_events: should song_played / song_skipped events (still) be send
        """
        self.playerid = playerid
        self.state = STOP
        self.song = None
        self.time = 0
        self.crossfade = False
        self.send_played_skipped_events = send_played_skipped_events

    def __cmp__(self, other):
        return (cmp(self.playerid, other.playerid) or
                cmp(self.state, other.state) or
                cmp(self.song, other.song) or
                cmp(self.time, other.time) or
                cmp(self.crossfade, other.crossfade) or 
                cmp(self.send_played_skipped_events, other.send_played_skipped_events))

    def __str__(self):
        s = "player %s " % `self.playerid`
        if self.state==STOP:
            s = s + "stopped"
        elif self.state==PAUSE:
            s = s + "paused"
        elif self.state==PLAY:
            s = s + "playing"
        s = s + " song: "
        if self.song:
            s = s + "%r at time %f" % ( self.song, self.time)
        else:
            s = s + "None"
        if self.crossfade:
            s = s+ " (crossfading)"
        return s

    def updatesong(self, song):
        """ update song and reset time """
        self.song = song
        self.time = 0

    def stopped(self):
        self.state = STOP
        self.song = None
        self.time = 0
        self.crossfade = False

    def paused(self):
        self.state = PAUSE

    def playing(self):
        self.state = PLAY

    def updatetime(self, time):
        self.time = time

    def updatecrossfade(self, crossfade):
        self.crossfade = crossfade

    def isplaying(self):
        return self.state == PLAY

    def ispaused(self):
        return self.state == PAUSE

    def isstopped(self):
        return self.state == STOP

    def iscrossfading(self):
        return self.crossfade and not self.state == STOP


class genericplayer(service.service):
    def __init__(self, id, playlistid, autoplay):
        """create a new player

        id:         the player id
        playlistid: playlist responsible for feeding player with songs. Set to None, if
                    there is no playlist for the player.
        autoplay:   should the player start automatically, if a song is in the playlist
                    and it has not been stopped explicitely by the user

        """
        service.service.__init__(self, "player %s" % id, daemonize=True)
        self.id = id
        self.autoplay = autoplay
        self.playlistid = playlistid

        # if wantplay != autoplay, the user has requested a player stop and thus
        # autoplay is effectively turned off, until the player is restarted again
        self.wantplay = autoplay

        # should we notify the database that the song has been played
        self.sendplayedevent = False

        # the playbackinfo structure describes the current player state
        # If a playlist is attached to the player, we issue song_played and song_skipped events
        self.playbackinfo = playbackinfo(self.id, playlistid is not None)

        # old playbackinfo, used to detect changes of the player state
        self.oplaybackinfo = copy.copy(self.playbackinfo)

        self.channel.subscribe(events.playerstart, self.playerstart)
        self.channel.subscribe(events.playerpause, self.playerpause)
        self.channel.subscribe(events.playertogglepause, self.playertogglepause)
        self.channel.subscribe(events.playerstop, self.playerstop)
        self.channel.subscribe(events.playernext, self.playernext)
        self.channel.subscribe(events.playerprevious, self.playerprevious)
        self.channel.subscribe(events.playerseekrelative, self.playerseekrelative)
        self.channel.subscribe(events.playerplaysong, self.playerplaysong)
        self.channel.subscribe(events.playerratecurrentsong, self.playerratecurrentsong)
        self.channel.subscribe(events.player_change_volume_relative, self.player_change_volume_relative)
        self.channel.subscribe(events.playerplayfaster, self.playerplayfaster)
        self.channel.subscribe(events.playerplayslower, self.playerplayslower)
        self.channel.subscribe(events.playerspeedreset, self.playerspeedreset)

        self.channel.supply(requests.getplaybackinfo, self.getplaybackinfo)

    def work(self):
        if self.isplaying():
            if ( self.playbackinfo.send_played_skipped_events and
                 self.playbackinfo.song and (self.playbackinfo.song.length < 10 or self.playbackinfo.time > 0.8*self.playbackinfo.song.length) ):
                song = self.playbackinfo.song
                hub.notify(events.song_played(song.songdbid, song, time.time()-self.playbackinfo.time))
                self.playbackinfo.send_played_skipped_events = False
            self.play()

        # request a new song, if none is playing and the player wants to play
        if self.isstopped() and self.wantplay:
            self.requestnextsong()

        # process incoming events
        self.channel.process()

        # and notify the rest of any changes in the playback status
        self.updatestatus()
        # Now the queue of all pending events has been
        # cleared. Depending on the player status we can now wait for
        # further incoming events.
        if not self.isplaying():
            # We sleep a little bit to prevent being overly active
            # when the event channel is spilled by messages
            time.sleep(0.2)
            # In this case, we can safely block since we will be waked
            # up by any message on the event channel. Thus, event if
            # we want to request a new song, we can rely on an event
            # signaling the addition of a new song to the playlist.
            # Before doing so, we release the player device
            self._playerreleasedevice()
            self.channel.process(block=True)

    def play(self):
        """play songs

        this method has to be implemented by specialized classes"""
        pass

    def updatestatus(self):
        """notify interested parties of changes in player status"""
        if self.oplaybackinfo != self.playbackinfo:
            self.oplaybackinfo = copy.copy(self.playbackinfo)
            hub.notify(events.playbackinfochanged(self.playbackinfo))

    def requestnextsong(self, manual=False, previous=False):
        """request next song from playlist and play it"""
        if self.playlistid is not None:
            nextsong = hub.request(requests.playlist_requestnextsong(self.playlistid, previous))
            self.playsong(nextsong, manual)

    def playsong(self, song, manual):
        """add song to playlist and mark song played, if song is not None

        manual indicates whether the user has requested the song manually
        """
        if song:
            self._playsong(song, manual)
            self.playbackinfo.playing()

    def isstopped(self):
        return self.playbackinfo.state == STOP

    def ispaused(self):
        return self.playbackinfo.state == PAUSE

    def isplaying(self):
        return self.playbackinfo.state == PLAY

    def _playsong(self, song, manual):
        """add song to playlist

        manual indicates whether the user has requested the song manually

        this method has to be implemented by specialized classes"""
        pass

    def _playerstart(self):
        """prepare player for playing

        this method has to be implemented by specialized classes"""
        pass

    def _playerpause(self):
        """pause player 

        this method has to be implemented by specialized classes"""
        pass

    def _playerunpause(self):
        """restart player after pause

        this method has to be implemented by specialized classes"""
        pass

    def _playerstop(self):
        """stop playing

        this method has to be implemented by specialized classes"""
        pass
    
    def _playerseekrelative(self, seconds):
        """seek by the given number of seconds in file (relative to current position)

        this method has to be implemented by specialized classes"""
        pass

    def _player_change_volume_relative(self, volume_adj):
        """change playback volume by volume_adj percent

        this method has to be implemented by specialized classes"""
        pass


    def _playerplayfaster(self):
        """Speed up the play rate of the current song
        this method has to be implemented by specialized classes"""
        pass

    def _playerplayslower(self):
        """Slow down the play rate of the current song
        this method has to be implemented by specialized classes"""
        pass

    def _playerspeedreset(self):
        """Reset the play rate of the current song back to normal
        this method has to be implemented by specialized classes"""
        pass

    def _playerreleasedevice(self):
        """temporarily release audio device 

        this method has to be implemented by specialized classes"""
        pass

    def _playerquit(self):
        """quit player

        this method has to be implemented by specialized classes"""
        pass

    # event handlers

    def playerstart(self, event):
        """start playing"""
        if event.playerid == self.id:
            if self.ispaused():
                self._playerunpause()
            elif self.isstopped():
                self.wantplay = self.autoplay
                self._playerstart()
                self.requestnextsong()
            self.playbackinfo.playing()

    def playerpause(self, event):
        """start/pause player"""
        if event.playerid == self.id:
            if self.isplaying():
                self.playbackinfo.paused()
                self._playerpause()

    def playertogglepause(self, event):
        """start/pause player"""
        if event.playerid == self.id:
            if self.isplaying():
                self.playbackinfo.paused()
                self._playerpause()
            elif self.ispaused():
                self.playbackinfo.playing()
                self._playerunpause()

    def playerstop(self, event):
        """stop playing"""
        if event.playerid == self.id:
            self.wantplay = False
            self.playbackinfo.stopped()
            self._playerstop()

    def playernext(self, event):
        """immediately play next song"""
        if event.playerid == self.id:
            # mark playing of song as skipped if it belongs to a playlist
            if self.playbackinfo.send_played_skipped_events and self.playbackinfo.song:
                song = self.playbackinfo.song
                hub.notify(events.song_skipped(song.songdbid, song))
                # we also prevent this song from being registered as played
                self.playbackinfo.send_played_skipped_events = False
            self.requestnextsong(manual=1)

    def playerprevious(self, event):
        """immediately play previous song"""
        if event.playerid == self.id:
            self.requestnextsong(manual=1, previous=1)

    def playerseekrelative(self, event):
        """seek by event.seconds in file (relative to current position """
        if event.playerid == self.id:
            self._playerseekrelative(event.seconds)

    def player_change_volume_relative(self, event):
        """ change volume relative to given setting """
        if event.playerid == self.id:
            self._player_change_volume_relative(event.volume_adj)

    def playerplayfaster(self, event):
        if event.playerid == self.id:
            self._playerplayfaster()

    def playerplayslower(self, event):
        if event.playerid == self.id:
            self._playerplayslower()

    def playerspeedreset(self, event):
        if event.playerid == self.id:
            self._playerspeedreset()

    def playerplaysong(self, event):
        """play event.song next"""
        if event.playerid == self.id:
            self.playsong(event.playlistitemorsong, manual=1)

    def playerratecurrentsong(self, event):
        """play event.song next"""
        if event.playerid == self.id and self.playbackinfo.song and 1 <= event.rating <= 5:
            self.playbackinfo.song.rate(event.rating)

    def quit(self, event):
        """quit player"""
        service.service.quit(self, event)
        self._playerquit()

    # request handlers

    def getplaybackinfo(self, request):
        if self.id != request.playerid:
            raise hub.DenyRequest
        else:
            return self.playbackinfo



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import events, hub
import statusbar
import version
import window

# a scrollable window which supports automatic closing 

class messagewin(window.window):

    """ a scrollable window which supports automatic closing """

    def __init__(self, screen, maxh, maxw, channel, colors, title, items, autoclosetime):
        self.maxh = maxh
        self.maxw = maxw
        self.channel = channel
        self.autoclosetime = autoclosetime
        self.items = items
        self.first = 0

        # for identification purposes, we only generate this once
        self.hidewindowevent = events.hidewindow(self)
        
        self.keybindings = config.keybindings.filelistwindow

        window.window.__init__(self, screen, 1, 1, 0, 0, colors, title)

        self.channel.subscribe(events.keypressed, self.keypressed)
        self.channel.subscribe(events.mouseevent, self.mouseevent)
        self.channel.subscribe(events.hidewindow, self.hidewindow)
        self.channel.subscribe(events.focuschanged, self.focuschanged)

    def _outputlen(self, width):
        """number of lines in window"""
        return len(self.items)

    def _resize(self):
        if self.maxw<=80:
            width = self.maxw
        else:
            width = 80 + int((self.maxw-80)*0.8)
        height = min(self._outputlen(width-2)+2, self.maxh-3)
        y = max(0, (self.maxh-height)/2)
        x = max(0, (self.maxw-width)/2)
        window.window.resize(self, height, width, y, x)

    def resize(self, maxh, maxw):
        self.maxh = maxh
        self.maxw = maxw
        self._resize()
        if self.hasfocus():
            self.update()

    def show(self):
        self._resize()
        self.first = 0
        self.top()
        self.update()
        if self.autoclosetime:
            hub.notify(events.sendeventin(self.hidewindowevent,
                                              self.autoclosetime,
                                              replace=1))

    def showitems(self):
        """ has to be implemented by derivec classes """
        pass
    
    # event handler

    def keypressed(self, event):
        if self.hasfocus():
            key = event.key
            # XXX: do we need own keybindings for help?
            if key in self.keybindings["selectnext"]:
                self.first = min(self._outputlen(self.iw)-self.ih, self.first+1)
            elif key in self.keybindings["selectprev"]:
                self.first = max(0, self.first-1)
            elif key in self.keybindings["selectnextpage"]:
                self.first = min(self._outputlen(self.iw)-self.ih, self.first+self.ih)
            elif key in self.keybindings["selectprevpage"]:
                self.first = max(0, self.first-self.iw)
            elif key in self.keybindings["selectfirst"]:
                self.first = 0
            elif key in self.keybindings["selectlast"]:
                self.first = self._outputlen(self.iw)-self.ih
            else:
                self.hide()
                raise hub.TerminateEventProcessing
                return

            self.update()
            if self.autoclosetime:
                hub.notify(events.sendeventin(self.hidewindowevent,
                                                  self.autoclosetime,
                                                  replace=1))
        
            raise hub.TerminateEventProcessing

    def mouseevent(self, event):
        if self.hasfocus():
            self.hide()
            raise hub.TerminateEventProcessing

    def focuschanged(self, event):
        if self.hasfocus():
            sbar = [("PyTone %s" % version.version, config.colors.statusbar.key)]
            sbar += statusbar.separator
            sbar += [(version.copyright, config.colors.statusbar.description)]
            sbar += statusbar.terminate
            hub.notify(events.statusbar_update(0, sbar))

    def update(self):
        window.window.update(self)
        self.showitems()






# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import copy

import events, hub, requests
import log
import network
import service

#
# songdb class
#

class songdb(service.service):
    def __init__(self, id, networklocation, songdbhub):
        service.service.__init__(self, "remote songdb", hub=songdbhub)
        self.id = id
        self.networklocation = networklocation
        self.remotesongdbid = "main"
        
        self.networkchannel = network.clientchannel(self.networklocation)
        #self.networkchannel.transmit(events.updatesong)
        #self.networkchannel.transmit(events.updatealbum)
        #self.networkchannel.transmit(events.updateartist)
        #self.networkchannel.transmit(events.playlistaddsong)
        self.networkchannel.start()

        # we need to be informed about database changes
        #self.channel.subscribe(events.updatesong, self.updatesong)
        #self.channel.subscribe(events.updatealbum, self.updatealbum)
        #self.channel.subscribe(events.updateartist, self.updateartist)
        #self.channel.subscribe(events.registersongs, self.registersongs)
        #self.channel.subscribe(events.registerplaylists, self.registerplaylists)
        
        # we are a database service provider...
        self.channel.supply(requests.dbrequest, self.dbrequest)
        log.info(_("database %s: type remote, location: %s") % (self.id,
                                                                self.networklocation))

    # request handler

    def dbrequest(self, request):
        if self.id != request.songdbid:
            raise hub.DenyRequest
        log.debug("dispatching %s" % `request`)
        # we have to copy the request, because another thread may also access it
        request = copy.copy(request)
        request.songdbid = self.remotesongdbid
        result = self.networkchannel.request(request)
        log.debug("result %s" % `result`)
        # we change the databasestats accordingly
        if isinstance(request, requests.getdatabasestats):
            result.type = "remote"
            result.location = str(self.networklocation)
        return result



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import events, requests, hub
import item
import slist
import log

class filelist(slist.slist):

    def __init__(self, win, songdbids):
        slist.slist.__init__(self, win, config.filelistwindow.scrollmode == "page")

        self.basedir = item.basedir(songdbids, rootdir=True)
        # self.basedir = item.basedir(songdbids)
        self.dir = [self.basedir]
        self.shistory = []
        self.readdir()

        self.win.channel.subscribe(events.songschanged, self.songschanged)
        self.win.channel.subscribe(events.artistschanged, self.artistschanged)
        self.win.channel.subscribe(events.albumschanged, self.albumschanged)
        self.win.channel.subscribe(events.tagschanged, self.tagschanged)
        self.win.channel.subscribe(events.songchanged, self.songchanged)
        self.win.channel.subscribe(events.dbplaylistchanged, self.dbplaylistchanged)
        self.win.channel.subscribe(events.filelistjumptosong, self.filelistjumptosong)

    def isdirselected(self):
        return isinstance(self.getselected(), item.diritem)

    def issongselected(self):
        return isinstance(self.getselected(), item.song)

    def getselectedsubdir(self):
        return self.dir + [self.getselected()]

    def readdir(self):
        self.set(self.dir[-1].getcontents())

    def updatedir(self):
        """ reread directory trying to keep the current selection """
        self.set(self.dir[-1].getcontents(), keepselection=True)

    def dirdown(self):
        self.shistory.append((self.dir, self.selected, self.top))
        self.dir = self.getselectedsubdir()
        self.readdir()
        # In the case of the selected item having been an artist check
        # whether only one album is present. If yes directly jump to
        # this album.
        if config.filelistwindow.skipsinglealbums and isinstance(self.dir[-1], item.artist) and len(self) <= 2:
            self.dir = self.getselectedsubdir()
            self.readdir()

    def dirup(self):
        if len(self.shistory)>0:
            dir, selected, top = self.shistory.pop()
            self.dir = dir
            self.readdir()
            self.selected = selected
            self.top = top
            # the window size could have changed in the meantime, so we have to update top
            self._updatetop()
            self._notifyselectionchanged()

    def focus_on(self, searchstring):
        # remove any previous focus
        if isinstance(self.dir[-1], item.focus_on):
            self.dirup()
        self.shistory.append((self.dir, self.selected, self.top))
        songdbid = self.dir[-1].songdbid
        # if filters apply, use them
        try:
            filters = self.dir[-1].filters
        except AttributeError:
            filters = None
        self.dir = self.dir + [item.focus_on(songdbid, searchstring, filters)]
        self.readdir()

    def selectionpath(self):
        return self.dir[-1].getheader(self.getselected())

    def insertrecursiveselection(self):
        if self.isdirselected():
            songs = self.getselected().getcontentsrecursivesorted()
            hub.notify(events.playlistaddsongs(songs))
        elif self.issongselected():
            hub.notify(events.playlistaddsongs([self.getselected()]))

    def randominsertrecursiveselection(self):
        if self.isdirselected():
            songs = self.getselected().getcontentsrecursiverandom()
            hub.notify(events.playlistaddsongs(songs))
        elif self.issongselected():
            hub.notify(events.playlistaddsongs([self.getselected()]))

    def rateselection(self, rating):
        if self.isdirselected():
            if not isinstance(self.getselected(), (item.artist, item.album)):
                self.win.sendmessage(_("Not rating virtual directories!"))
                return False
            songs = self.getselected().getcontentsrecursive()
            if rating:
                self.win.sendmessage(_("Rating %d song(s) with %d star(s)...") % (len(songs), rating))
            else:
                self.win.sendmessage(_("Removing rating of %d song(s)...") % len(songs))
        elif self.issongselected():
            songs = [self.getselected()]
        for song in songs:
            song.rate(rating)
        return True

    def addtagselection(self, tag):
        if self.isdirselected():
            if not isinstance(self.getselected(), (item.artist, item.album)):
                self.win.sendmessage(_("Not tagging virtual directories!"))
                return False
            songs = self.getselected().getcontentsrecursive()
            self.win.sendmessage(_("Tagging %d song(s) with tag '%s'...") % (len(songs), tag))
        elif self.issongselected():
            songs = [self.getselected()]
        for song in songs:
            song.addtag(tag)
        return True

    def removetagselection(self, tag):
        if self.isdirselected():
            if not isinstance(self.getselected(), (item.artist, item.album)):
                self.win.sendmessage(_("Not untagging virtual directories!"))
                return False
            songs = self.getselected().getcontentsrecursive()
            self.win.sendmessage(_("Removing tag '%s' from %d song(s)...") % (tag, len(songs)))
        elif self.issongselected():
            songs = [self.getselected()]
        for song in songs:
            song.removetag(tag)
        return True

    def toggledeleteselection(self):
        if self.isdirselected():
            if not isinstance(self.getselected(), (item.artist, item.album)):
                self.win.sendmessage(_("Not (un)deleting virtual directories!"))
                return False
            songs = self.getselected().getcontentsrecursive()
            self.win.sendmessage(_("(Un)deleting %d song(s)...") % len(songs))
        elif self.issongselected():
            songs = [self.getselected()]
        for song in songs:
            song.toggledelete()
        self.updatedir()
        return True

    def rescanselection(self, force):
        if ( isinstance(self.getselected(), item.basedir) or
             ( isinstance(self.getselected(), item.filesystemdir) and self.getselected().isbasedir()) ):
            # instead of rescanning of a whole filesystem we start the autoregisterer
            self.win.sendmessage(_("Scanning for songs in database '%s'...") % self.getselected().songdbid)
            hub.notify(events.autoregistersongs(self.getselected().songdbid))
        else:
            if self.isdirselected():
                # distribute songs over songdbs
                # Note that we have to ensure that only dbitem.song (and not item.song) instances
                # are sent to the db
                songs = self.getselected().getcontentsrecursive()
            else:
                songs = [self.getselected()]
            self.win.sendmessage(_("Rescanning %d song(s)...") % len(songs))
            dsongs = {}
            for song in songs:
                dsongs.setdefault(song.songdbid, []).append(song)
            for songdbid, songs in dsongs.items():
                if songs:
                    hub.notify(events.autoregisterer_rescansongs(songdbid, songs, force))

    # event handler

    def songschanged(self, event):
        if isinstance( self.dir[-1], (item.songs, item.album)):
            self.updatedir()
            self.win.update()

    def artistschanged(self, event):
        if isinstance( self.dir[-1], item.basedir):
            self.updatedir()
            self.win.update()

    def albumschanged(self, event):
        if isinstance(self.dir[-1], (item.albums, item.artist, item.compilations)):
            self.updatedir()
            self.win.update()

    def tagschanged(self, event):
        if isinstance(self.dir[-1], item.tags):
            self.updatedir()
            self.win.update()

    def songchanged(self, event):
        if isinstance( self.dir[-1], (item.songs, item.album, item.topplayedsongs, item.lastplayedsongs)):
            self.updatedir()
            self.win.update()

    def dbplaylistchanged(self, event):
        #if (isinstance(self.dir[-1], item.artist) and
        #    self.dir[-1].songdbid==event.songdbid and
        #    self.dir[-1].name in event.album.artists):
        self.updatedir()
        self.win.update()

    def filelistjumptosong(self, event):
        """ directly jump to given song """
        # In order to get the correct shistory, we more or less simulate
        # a walk through the directory hierarchy, starting from the basedir.
        self.shistory = []
        self.dir = [self.basedir]
        self.readdir()
        # either we are able to locate the artist or we should look under compilations
        if ( (event.song.artist_id and (self.selectbyid(event.song.artist_id) or self.selectbyid("compilations"))) or
             self.selectbyid("noartist") ):
            self.dirdown()
            # We might have skipped the album when there is only a single one of
            # the given artist.
            if not isinstance(self.dir[-1], item.album) and self.selectbyid(event.song.album_id):
                self.dirdown()
            self.selectbyid(event.song.id)
        self.win.update()
